import time
import numpy as np
import capytaine as cpt
import xarray as xr
import torch

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

reference_cpu_solver = cpt.BEMSolver()


# A simple implementation of a linear solver on GPU using PyTorch
def linear_solver_on_GPU(A, b):
    A_on_GPU = torch.tensor(np.array(A), dtype=torch.complex64, device=device)
    b_on_GPU = torch.tensor(b, dtype=torch.complex64, device=device)
    res = torch.linalg.solve(A_on_GPU, b_on_GPU)
    return res.cpu().numpy()

naive_gpu_solver = cpt.BEMSolver(
    engine=cpt.BasicMatrixEngine(linear_solver=linear_solver_on_GPU)
)


# Alternatively, an optimization of the above to avoid doing the same GPU
# transfer and LU factorization twice
latest_A = None
latest_LU_decomp = None
def lu_linear_solver_with_cache_on_GPU(A, b):
    global latest_A, latest_LU_decomp
    if A is not latest_A:
        latest_A = A
        A_on_GPU = torch.tensor(np.array(A), dtype=torch.complex64, device=device)
        latest_LU_decomp = torch.linalg.lu_factor(A_on_GPU)
    b_on_GPU = torch.tensor(np.array(b), dtype=torch.complex64, device=device).reshape(-1, 1)
    # Reshaping because `torch.linalg.lu_solve` wants a column vector on the
    # right-hand-side (unlike `torch.linalg.solve`).
    res = torch.linalg.lu_solve(*latest_LU_decomp, b_on_GPU).reshape(-1)
    return res.cpu().numpy()

gpu_solver = cpt.BEMSolver(
    engine=cpt.BasicMatrixEngine(linear_solver=lu_linear_solver_with_cache_on_GPU)
)


###############
#  Benchmark  #
###############

mesh = cpt.mesh_sphere(resolution=(50, 50)).immersed_part()
print(mesh.nb_faces, "faces")
body = cpt.FloatingBody(mesh, dofs=cpt.rigid_body_dofs())

test_matrix = xr.Dataset(
    coords={
        "omega": np.linspace(0.0, 4.0, 20),
        "radiating_dof": list(body.dofs),
        "water_depth": [np.inf],
    }
)

start = time.perf_counter()
ds1 = reference_cpu_solver.fill_dataset(test_matrix, body)
print("CPU:", time.perf_counter() - start)

start = time.perf_counter()
ds2 = naive_gpu_solver.fill_dataset(test_matrix, body)
print("GPU:", time.perf_counter() - start)

start = time.perf_counter()
ds3 = gpu_solver.fill_dataset(test_matrix, body)
print("Slightly faster GPU:", time.perf_counter() - start)


# Check outputs
import matplotlib.pyplot as plt

fig, ax = plt.subplots()
ax.plot(
    ds1.omega.values,
    ds1.added_mass.sel(radiating_dof="Roll", influenced_dof="Roll").values,
)
ax.plot(
    ds2.omega.values,
    ds2.added_mass.sel(radiating_dof="Roll", influenced_dof="Roll").values,
)
ax.plot(
    ds3.omega.values,
    ds3.added_mass.sel(radiating_dof="Roll", influenced_dof="Roll").values,
)
plt.show()



import numpy as np
import capytaine as cpt
import matplotlib.pyplot as plt

mesh = cpt.mesh_sphere(faces_max_radius=0.1).immersed_part()
body = cpt.FloatingBody(
    mesh=mesh,
    dofs=cpt.rigid_body_dofs(rotation_center=(0, 0, 0))
)

pb = cpt.RadiationProblem(wavelength=1.0, body=body, radiating_dof="Surge", water_depth=10.0)
solver = cpt.BEMSolver()
res = solver.solve(pb, keep_details=True)

body.show_matplotlib(
        color_field=np.real(res.pressure),
        cmap=plt.get_cmap("viridis"),  # Colormap
        )



import pytest

import numpy as np
import capytaine as cpt

##################
#  Generate lid  #
##################

def test_lid_below_free_surface():
    mesh = cpt.AxialSymmetricMesh.from_profile(lambda z: (z + 1.0)**2, np.linspace(-1.0, 0.0, 10)).merged()
    lid_mesh = mesh.generate_lid(z=-0.5)
    x, y, z = lid_mesh.faces_centers.T
    assert np.all(np.hypot(x, y) <= (z + 1.0)**2)


def test_lid_below_body():
    mesh = cpt.mesh_sphere(radius=0.5, center=(0, 0, 0.0))
    lid_mesh = mesh.generate_lid(z=-2.0)
    assert lid_mesh.nb_vertices == 0


def test_lid_underwater_mesh():
    mesh = cpt.mesh_sphere(radius=0.5, center=(0, 0, -1.5))
    lid_mesh = mesh.generate_lid()
    assert lid_mesh.nb_vertices == 0


def test_lid_concave_body():
    pygmsh = pytest.importorskip("pygmsh")
    from capytaine.io.meshio import load_from_meshio
    d = 1.9
    with pygmsh.occ.Geometry() as geom:
        cyl1 = geom.add_cylinder([0, 0, 0], [0, 0, -1.0],  1.0)
        cyl2 = geom.add_cylinder([d, 0, 0], [0.0, 0, -1.0],  1.0)
        geom.boolean_union([cyl1, cyl2])
        gmsh_mesh = geom.generate_mesh(dim=2)
    mesh = load_from_meshio(gmsh_mesh)
    lid_mesh = mesh.generate_lid()
    def in_crown(x, y):
        return (np.hypot(x, y) < 1.0) | (np.hypot(x-d, y) < 1.0)
    assert all(in_crown(lid_mesh.faces_centers[:, 0], lid_mesh.faces_centers[:, 1]))


def test_lid_non_simply_connected_crown():
    pygmsh = pytest.importorskip("pygmsh")
    from capytaine.io.meshio import load_from_meshio
    with pygmsh.occ.Geometry() as geom:
        geom.add_torus((0, 0, 0), 2.0, 0.5, mesh_size=0.3)
        gmsh_mesh = geom.generate_mesh(dim=2)
    mesh = load_from_meshio(gmsh_mesh).immersed_part().heal_mesh()
    lid_mesh = mesh.generate_lid()
    def in_crown(x, y):
        return (np.hypot(x, y) < 2.5) & (np.hypot(x, y) > 1.5)
    assert all(in_crown(lid_mesh.faces_centers[:, 0], lid_mesh.faces_centers[:, 1]))


def test_lid_non_connex_crown():
    pygmsh = pytest.importorskip("pygmsh")
    from capytaine.io.meshio import load_from_meshio
    with pygmsh.occ.Geometry() as geom:
        geom.add_torus((0, 0, 0), 2.0, 0.5, mesh_size=0.4)
        gmsh_mesh = geom.generate_mesh(dim=2)
    mesh = load_from_meshio(gmsh_mesh).rotated_x(np.pi/2).immersed_part().heal_mesh()
    lid_mesh = mesh.generate_lid()
    def in_crown(x, y):
        return (np.hypot(x-2.0, y) < 0.5) | (np.hypot(x+2.0, y) < 0.5)
    assert all(in_crown(lid_mesh.faces_centers[:, 0], lid_mesh.faces_centers[:, 1]))


def test_lid_multibody():
    mesh = cpt.mesh_sphere(center=(0, 0, 0)) + cpt.mesh_sphere(center=(0, 5, 0))
    mesh = mesh.immersed_part()
    lid_mesh = mesh.generate_lid()
    def in_crown(x, y):
        return (np.hypot(x, y) < 1.0) | (np.hypot(x, y-5) < 1.0)
    assert all(in_crown(lid_mesh.faces_centers[:, 0], lid_mesh.faces_centers[:, 1]))


def test_lid_symmetric_mesh():
    mesh = cpt.mesh_vertical_cylinder(radius=1, reflection_symmetry=True).immersed_part()
    lid_mesh = mesh.generate_lid()
    assert isinstance(lid_mesh, cpt.ReflectionSymmetricMesh)
    def in_crown(x, y):
        return np.hypot(x, y) < 1.0
    assert all(in_crown(lid_mesh.faces_centers[:, 0], lid_mesh.faces_centers[:, 1]))


def test_lid_nested_symmetric_mesh():
    mesh = cpt.mesh_parallelepiped(reflection_symmetry=True).immersed_part()
    lid_mesh = mesh.generate_lid()
    assert isinstance(lid_mesh, cpt.ReflectionSymmetricMesh)
    assert isinstance(lid_mesh.half, cpt.ReflectionSymmetricMesh)
    def in_crown(x, y):
        return np.maximum(np.abs(x), np.abs(y)) < 0.5
    assert all(in_crown(lid_mesh.faces_centers[:, 0], lid_mesh.faces_centers[:, 1]))


def test_lid_auto_position():
    mesh = cpt.mesh_vertical_cylinder(radius=1.0).immersed_part()
    lid_mesh = mesh.generate_lid(z=mesh.lowest_lid_position(omega_max=5.0))
    body = cpt.FloatingBody(mesh=mesh, lid_mesh=lid_mesh)
    assert body.first_irregular_frequency_estimate() > 5.0


def test_lid_immersed_part():
    mesh = cpt.mesh_vertical_cylinder(radius=1.0).immersed_part()
    lid_mesh = mesh.generate_lid(z=0.0)
    assert lid_mesh.immersed_part().nb_faces > 0
    assert lid_mesh.immersed_part() == lid_mesh


def test_clipped_lid_above_the_free_surface():
    mesh = cpt.mesh_vertical_cylinder(radius=1.0).immersed_part()
    lid_mesh = mesh.generate_lid(z=0.5).immersed_part()
    body = cpt.FloatingBody(mesh=mesh, lid_mesh=lid_mesh)
    assert body.lid_mesh is None


def test_clipped_lid_above_the_free_surface__():
    mesh = cpt.mesh_vertical_cylinder(radius=1.0).immersed_part()
    lid_mesh = mesh.generate_lid(z=0.5)
    body = cpt.FloatingBody(mesh=mesh, lid_mesh=lid_mesh)
    body = body.immersed_part()
    assert body.lid_mesh is None


def test_lid_on_vertical_panels_only():
    # Used to raise an error.
    # https://github.com/capytaine/capytaine/issues/625
    mesh = cpt.mesh_rectangle(normal=(1, 0, 0))
    lid_mesh = mesh.generate_lid()


####################
#  Lid extraction  #
####################

def test_extract_lid():
    mesh = cpt.mesh_vertical_cylinder(center=(0, 0, -2), length=4, resolution=(2, 8, 10))
    hull_mesh, lid_mesh = mesh.extract_lid()
    assert lid_mesh.nb_faces == 2*8


def test_extract_lid_symmetric_mesh():
    mesh = cpt.mesh_vertical_cylinder(center=(0, 0, -2), length=4, resolution=(2, 8, 10), reflection_symmetry=True)
    hull_mesh, lid_mesh = mesh.extract_lid()
    assert isinstance(hull_mesh, cpt.ReflectionSymmetricMesh)
    assert isinstance(lid_mesh, cpt.ReflectionSymmetricMesh)


def test_extract_lid_nested_symmetric_mesh():
    mesh = cpt.mesh_parallelepiped(center=(0, 0, -1), reflection_symmetry=True)
    hull_mesh, lid_mesh = mesh.extract_lid()
    assert isinstance(hull_mesh, cpt.ReflectionSymmetricMesh)
    assert isinstance(hull_mesh.half, cpt.ReflectionSymmetricMesh)
    assert isinstance(lid_mesh, cpt.ReflectionSymmetricMesh)
    assert isinstance(lid_mesh.half, cpt.ReflectionSymmetricMesh)



import numpy as np
from numpy import dot, pi
from numpy.linalg import norm

import capytaine as cpt
from capytaine.green_functions.abstract_green_function import AbstractGreenFunction


class MyGreenFunction(AbstractGreenFunction):
    """An example of a custom routine to evaluate the Green function."""

    floating_point_precision = "float64"  # Optional, could allow for RAM usage optimisations

    def evaluate(
        self,
        mesh1,
        mesh2,
        free_surface,
        water_depth,
        wavenumber,
        adjoint_double_layer=True,
        early_dot_product=True,
    ):
        """The main method that needs to be implemented in the class."""

        if not adjoint_double_layer:
            raise NotImplementedError(
                """MyGreenFunction only implements the adjoint double layer"""
            )

        if free_surface != np.inf and water_depth != np.inf:
            raise NotImplementedError(
                """MyGreenFunction only implements the case without a free surface"""
            )

        # Initialize the matrices
        S = np.zeros((mesh1.nb_faces, mesh2.nb_faces))
        if early_dot_product:
            gradient_shape = (mesh1.nb_faces, mesh2.nb_faces)
        else:
            gradient_shape = (mesh1.nb_faces, mesh2.nb_faces, 3)
        K = np.zeros(gradient_shape)

        for i in range(mesh1.nb_faces):
            for j in range(mesh2.nb_faces):
                area = mesh1.faces_areas[j]
                if i == j:
                    # Approximation of the integral of the 1/r Green
                    # function over the panel by taking the integral over a
                    # circle of same center and same area.
                    S[i, i] = -1 / (4 * pi) * 2 * np.sqrt(area * pi)
                    if mesh1 is mesh2:
                        if early_dot_product:
                            K[i, i] = 1 / 2
                        else:
                            K[i, i, :] = 1 / 2 * mesh1.faces_normals[i, :]
                    else:
                        if early_dot_product:
                            K[i, i] = 0.0
                        else:
                            K[i, i, :] = 0.0

                else:
                    # The integral on the face is computed using the value
                    # at the center of the face.
                    r = mesh1.faces_centers[i, :] - mesh2.faces_centers[j, :]

                    S[i, j] = -area / (4 * pi * norm(r))
                    gradient_rankine = area * r / (4 * pi * norm(r) ** 3)
                    if early_dot_product:
                        K[i, j] = dot(gradient_rankine, mesh1.faces_normals[i, :])
                    else:
                        K[i, j, :] = gradient_rankine

        return S, K


# Define a solver using the Green function above.
custom_solver = cpt.BEMSolver(green_function=MyGreenFunction())
default_solver = cpt.BEMSolver()  # for comparison

# Example of a problem
radius = 1.4
body = cpt.FloatingBody(
    mesh=cpt.mesh_sphere(radius=radius, resolution=(20, 20)),
    dofs=cpt.rigid_body_dofs(rotation_center=(0, 0, 0)),
)
problem = cpt.RadiationProblem(
    body=body, free_surface=np.inf, water_depth=np.inf, radiating_dof="Surge"
)

custom_result = custom_solver.solve(problem)
default_result = default_solver.solve(problem)

print("Added mass:")
print(f"    Analytical value: {2/3*pi*problem.rho*radius**3:.2e}")
print(f"    Custom solver:    {custom_result.added_masses['Surge']:.2e}")
print(f"    Default solver:   {default_result.added_masses['Surge']:.2e}")
# The analytical solution is known for this particular test case.
# The custom solver might be more accurate. That is probably a coincidence.
# Both should converge to the same solution when the mesh is refined.

print()
print("Timer:")
print(f"    Custom solver:  {custom_solver.timer['  Green function'].mean:.1e} s")
print(f"    Default solver: {default_solver.timer['  Green function'].mean:.1e} s")



"""Tests for the mesh import of mesh from external libraries"""

import os
import pytest

import numpy as np
import xarray as xr

from capytaine.io.mesh_writers import write_STL
from capytaine.io.mesh_loaders import load_STL
import capytaine as cpt

#################################################################

def generate_pygmsh_wavebot():
    pygmsh = pytest.importorskip("pygmsh", reason="PyGMSH not installed, test skipped")
    offset = 1e-2 # small offset so you can clip at z=0
    T1 = 0.16
    T2 = 0.37
    r1 = 0.88
    r2 = 0.35
    with pygmsh.occ.Geometry() as geom:
        cyl = geom.add_cylinder([0,0,0], [0,0,-T1], r1)
        cone = geom.add_cone([0,0,-T1], [0,0,-T2], r1, r2)
        geom.translate(cyl, [0, 0, offset])
        geom.translate(cone, [0, 0, offset])
        geom.boolean_union([cyl, cone])
        mesh = geom.generate_mesh(dim=2)
    vol_exp = np.pi*T1*r1**2 + 1/3 * np.pi * T2 * (r2**2 + r2 * r1 + r1**2)
    return (mesh, vol_exp)

def generate_pygmsh_cylinder():
    pygmsh = pytest.importorskip("pygmsh", reason="PyGMSH not installed, test skipped")
    offset = 1e-2 # small offset so you can clip at z=0
    T = 0.52
    r1 = 0.88
    with pygmsh.occ.Geometry() as geom:
        cyl = geom.add_cylinder([0,0,0], [0,0,-T], r1)
        geom.translate(cyl, [0, 0, offset])
        mesh = geom.generate_mesh(dim=2)
    vol_exp = np.pi*r1**2*T
    return (mesh, vol_exp)

def generate_pygmsh_sphere():
    pygmsh = pytest.importorskip("pygmsh", reason="PyGMSH not installed, test skipped")
    offset = 1e-2 # small offset so you can clip at z=0
    r1 = 0.88
    with pygmsh.occ.Geometry() as geom:
        sphere = geom.add_ball(center=[0,0,0], radius=r1)
        geom.translate(sphere, [0, 0, offset])
        mesh = geom.generate_mesh(dim=2)
    vol_exp = 1/2*np.pi*4/3*r1**3
    return (mesh, vol_exp)


@pytest.mark.parametrize("generate_pygmsh",
                         [generate_pygmsh_cylinder, generate_pygmsh_sphere, generate_pygmsh_wavebot])
def test_from_meshio_pygmsh(generate_pygmsh):
    pytest.importorskip("pygmsh", reason="PyGMSH not installed, test skipped")
    mesh, vol_exp = generate_pygmsh()
    mesh = cpt.load_mesh(mesh)
    assert mesh.immersed_part().volume == pytest.approx(vol_exp, rel=1e-1)

    fb = cpt.FloatingBody(mesh.immersed_part())
    fb.add_translation_dof(name="Heave")
    test_matrix = xr.Dataset(coords={
        'omega': [np.pi],
        'wave_direction': [0],
        'radiating_dof': list(fb.dofs.keys()),
        })
    solver = cpt.BEMSolver()
    solver.fill_dataset(test_matrix, bodies=fb)

#################################################################

def test_STL(tmp_path):
    pytest.importorskip("vtk", reason="VTK not installed, test skipped.")
    mesh = cpt.mesh_sphere()
    filepath = tmp_path / "test.stl"
    write_STL(filepath, mesh.vertices, mesh.faces)
    reloaded_mesh = load_STL(str(filepath), name="Bla")

    assert reloaded_mesh.name == "Bla"
    assert np.allclose(mesh.vertices, reloaded_mesh.vertices)


def test_STL_with_uppercase_file_extension(tmp_path):
    pytest.importorskip("vtk", reason="VTK not installed, test skipped.")
    mesh = cpt.mesh_sphere()
    filepath = tmp_path / "test.STL"
    write_STL(filepath, mesh.vertices, mesh.faces)
    reloaded_mesh = cpt.load_mesh(str(filepath), name="Bla")
    # Should detect that the file is stl using the file extension

    assert reloaded_mesh.name == "Bla"
    assert np.allclose(mesh.vertices, reloaded_mesh.vertices)


def test_STL_with_meshio(tmp_path):
    pytest.importorskip("vtk", reason="VTK not installed, test skipped.")
    pytest.importorskip("meshio", reason="Meshio not installed, test skipped")
    from capytaine.io.meshio import load_from_meshio
    mesh, _ = generate_pygmsh_cylinder()
    cpt_mesh = load_from_meshio(mesh)
    cpt_mesh.heal_mesh()

    # Write with Meshio and reload with Meshmagick
    mesh.write(tmp_path / "wavebot.stl")
    cpt_mesh_2 = cpt.load_mesh(tmp_path / "wavebot.stl")
    assert np.allclose(cpt_mesh.vertices, cpt_mesh_2.vertices)

    # # Write with Meshmagick and reload with Meshio
    # # FAILING FOR NOW
    # # ERROR IN STL LOADER IN MESHMAGICK
    # write_STL(tmp_path / "wavebot2.stl", fb.mesh.vertices, fb.mesh.faces)
    # mesh2 = meshio.read(tmp_path / "wavebot2.stl")
    # fb3 = cpt.FloatingBody.from_meshio(mesh2)
    # assert np.allclose(fb.mesh.vertices, fb3.mesh.vertices)

#################################################################

def test_MED_file():
    pytest.importorskip("h5py", reason="h5py not installed, test skipped")
    mesh = cpt.load_mesh(os.path.join(os.path.dirname(__file__), "mesh_files_examples/barge.med"))
    assert mesh.nb_faces == 187

def test_MSH2_file():
    mesh = cpt.load_mesh(os.path.join(os.path.dirname(__file__), "mesh_files_examples/cylinder2.msh"))
    assert mesh.nb_faces == 64

def test_MSH4_file():
    pytest.importorskip("meshio", reason="meshio not installed, test skipped")
    mesh = cpt.load_mesh(os.path.join(os.path.dirname(__file__), "mesh_files_examples/cylinder4.msh"))
    assert mesh.nb_faces == 64



import numpy as np
import logging
from typing import Union, Iterable, Tuple, TextIO
import xarray

logger = logging.getLogger(__name__)

DOF_INDEX = {"Surge": 1, "Sway": 2, "Heave": 3, "Roll": 4, "Pitch": 5, "Yaw": 6}

DOF_TYPE = {
    dof: "trans" if dof in {"Surge", "Sway", "Heave"} else "rot" for dof in DOF_INDEX
}
K_LOOKUP = {
    ("trans", "trans"): 3,
    ("trans", "rot"): 4,
    ("rot", "trans"): 4,
    ("rot", "rot"): 5,
}


def get_dof_index_and_k(dof_i: str, dof_j: str) -> Tuple[int, int, int]:
    """Get the degree of freedom indices and the corresponding stiffness matrix index.

    Parameters
    ----------
    dof_i: str
        The name of the first degree of freedom.
    dof_j: str
        The name of the second degree of freedom.

    Returns
    -------
    tuple
        A tuple containing the indices (i, j) and the stiffness matrix index k.
    """
    i = DOF_INDEX[dof_i]
    j = DOF_INDEX[dof_j]
    t_i = DOF_TYPE[dof_i]
    t_j = DOF_TYPE[dof_j]
    k = K_LOOKUP[(t_i, t_j)]
    return i, j, k


def check_dataset_ready_for_export(ds: xarray.Dataset) -> None:
    """
    Sanity checks to validate that the dataset is exportable to BEMIO/WAMIT-like formats.

    Parameters
    ----------
    ds : xarray.Dataset
        The dataset to be validated.

    Raises
    ------
    ValueError
        If any unsupported coordinate has multiple values or
        if non-rigid-body DOFs are present.
    """
    # 1. Check for singleton coordinates
    critical_coords = ["water_depth", "g", "rho"]
    coords_with_multiple_values = [
        k
        for k in critical_coords
        if k in ds.coords and len(ds.coords[k].dims) > 0 and ds.sizes[k] > 1
    ]

    if coords_with_multiple_values:
        msg = (
            "Export formats like WAMIT require only one value for each of the following coordinates: "
            f"{', '.join(critical_coords)}.\n"
            f"Problematic dimensions: {coords_with_multiple_values}.\n"
            "You can extract a subset using:\n"
            f"    ds_slice = ds.sel({', '.join([f'{k}={str(ds.coords[k].values[0])}' for k in coords_with_multiple_values])})"
        )
        raise ValueError(msg)

    # 2. Check for rigid-body DOFs only
    rigid_body_dofs = ("Surge", "Sway", "Heave", "Roll", "Pitch", "Yaw")
    if "influenced_dof" in ds.coords:
        dofs = set(ds.influenced_dof.values)
        non_rigid_dofs = dofs.difference(set(rigid_body_dofs))
        if non_rigid_dofs:
            raise ValueError(
                "WAMIT Export is only supported for single rigid body.\n"
                f"Unexpected DOFs: {non_rigid_dofs}.\n"
                f"Allowed DOFs: {rigid_body_dofs}"
            )


def identify_frequency_axis(
    dataset: Union[xarray.Dataset, xarray.DataArray],
) -> Tuple[str, np.ndarray, np.ndarray]:
    """
    Identify the frequency axis in the dataset and return its values along with the periods.

    Parameters
    ----------
    dataset : xarray.Dataset or xarray.DataArray
        Dataset that must include 'period' coordinate and at least one of:
        'omega', 'freq', 'period', 'wavenumber', or 'wavelength' as dimension.

    Returns
    -------
    freq_key : str
        The name of the main frequency-like coordinate.
    freq_vals : np.ndarray
        The values from the frequency coordinate (as present in the dataset).
    period_vals : np.ndarray
        The values of the 'period' coordinate in seconds.

    Raises
    ------
    ValueError
        If 'period' is not a coordinate or if no frequency-like dimension is found.
    """
    allowed_keys = {"omega", "freq", "wavenumber", "wavelength", "period"}
    dataset_dims = set(dataset.dims)
    keys_in_dataset = dataset_dims & allowed_keys

    if "period" not in dataset.coords:
        raise ValueError("Dataset must contain 'period' as a coordinate.")

    # Prioritize 'period' if it is one of the dimensions
    if "period" in dataset_dims:
        freq_key = "period"
    elif len(keys_in_dataset) >= 1:
        freq_key = sorted(keys_in_dataset)[0]  # deterministic choice
    else:
        raise ValueError(
            "Dataset must contain at least one frequency-like dimension among: "
            "'omega', 'freq', 'wavenumber', 'wavelength', or 'period'."
        )

    freq_vals = np.asarray(dataset[freq_key].values)
    period_vals = np.asarray(dataset["period"].values)

    return freq_key, freq_vals, period_vals


def export_wamit_hst(
    dataset: xarray.Dataset, filename: str, length_scale: float = 1.0
) -> None:
    """
    Export the nondimensional hydrostatic stiffness matrix to a WAMIT .hst file.

    Format:
        I     J     C(I,J)

    Parameters
    ----------
    dataset: xarray.Dataset
        Must contain 'hydrostatics' field with 'hydrostatic_stiffness' (named-dict or labeled 6x6 array).
        Must also contain 'rho' and 'g' (either in dataset or within hydrostatics).
    filename: str
        Output path for the .hst file.
    length_scale: float
        Reference length scale L for nondimensionalization.
    """
    if "hydrostatic_stiffness" not in dataset:
        raise ValueError("Dataset must contain a 'hydrostatic_stiffness' field.")

    # Reduce all extra dimensions to their first value, except the last two (should be 6x6)
    hydrostatic = dataset["hydrostatic_stiffness"]
    C = np.asarray(hydrostatic)
    if C is None or C.shape != (6, 6):
        raise ValueError("'hydrostatic_stiffness' must be a 6x6 matrix.")

    rho = float(np.atleast_1d(dataset.get("rho")).item())
    g = float(np.atleast_1d(dataset.get("g")).item())

    # DOF order used in Capytaine
    dof_names = ["Surge", "Sway", "Heave", "Roll", "Pitch", "Yaw"]

    with open(filename, "w") as f:
        for i_local, dof_i in enumerate(dof_names):
            for j_local, dof_j in enumerate(dof_names):
                cij = C[i_local, j_local]
                i, j, k = get_dof_index_and_k(dof_i, dof_j)
                norm = rho * g * (length_scale**k)
                cij_nd = cij / norm
                f.write(f"{i:5d} {j:5d} {cij_nd:12.6e}\n")


def export_wamit_1(
    dataset: xarray.Dataset, filename: str, length_scale: float = 1.0
) -> None:
    """
    Export added mass and radiation damping coefficients to a WAMIT .1 file.

    Coefficients are normalized as:
        Aij = Aij / (rho * length_scale^k)
        Bij = Bij / (omega * rho * length_scale^k)

    Format:
        PER     I     J     Aij [Bij]

    Special handling:
        - For PER = -1 (omega = inf), only Aij is written.
        - For PER = 0  (omega = 0), only Aij is written.

    Parameters
    ----------
    dataset: xarray.Dataset
        Must contain 'added_mass', 'radiation_damping', and either 'omega' or 'period'.
    filename: str
        Output .1 file.
    length_scale: float
        Reference length scale (L) used for normalization.

    Raises
    ------
    ValueError
        If required fields are missing or forward speed is not zero.
    """
    if "added_mass" not in dataset or "radiation_damping" not in dataset:
        raise ValueError("Missing 'added_mass' or 'radiation_damping' in dataset.")

    if not np.isclose(dataset["forward_speed"].item(), 0.0):
        raise ValueError("Forward speed must be zero for WAMIT export.")

    rho = dataset["rho"].item()
    added_mass = dataset["added_mass"]
    damping = dataset["radiation_damping"]
    omegas = dataset["omega"]
    dofs = list(added_mass.coords["influenced_dof"].values)

    # Determine main frequency coordinate
    freq_key, freq_vals, period_vals = identify_frequency_axis(dataset=added_mass)

    # Separate lines into blocks depending on period type
    period_blocks = {
        "T_zero": [],  # period == 0    → omega = inf  → PER = -1
        "T_inf": [],  # period == inf  → omega = 0    → PER = 0
        "T_regular": [],  # finite, non-zero periods
    }

    for omega, freq_val, period in zip(omegas, freq_vals, period_vals):
        for dof_i in dofs:
            for dof_j in dofs:
                j_dof, i_dof, k = get_dof_index_and_k(dof_i, dof_j)
                A = added_mass.sel(
                    {
                        freq_key: freq_val,
                        "influenced_dof": dof_i,
                        "radiating_dof": dof_j,
                    }
                ).item()
                norm = rho * (length_scale**k)
                A_norm = A / norm

                if np.isclose(period, 0.0):
                    # Case PER = -1 (omega = inf)
                    line = f"{-1.0:12.6e}\t{i_dof:5d}\t{j_dof:5d}\t{A_norm:12.6e}\n"
                    period_blocks["T_zero"].append(line)
                elif np.isinf(period):
                    # Case PER = 0 (omega = 0)
                    line = f"{0.0:12.6e}\t{i_dof:5d}\t{j_dof:5d}\t{A_norm:12.6e}\n"
                    period_blocks["T_inf"].append(line)
                else:
                    B = damping.sel(
                        {
                            freq_key: freq_val,
                            "influenced_dof": dof_i,
                            "radiating_dof": dof_j,
                        }
                    ).item()
                    B_norm = B / (omega * norm)
                    line = f"{period:12.6e}\t{i_dof:5d}\t{j_dof:5d}\t{A_norm:12.6e}\t{B_norm:12.6e}\n"
                    period_blocks["T_regular"].append((period, line))

    # Sort regular lines by increasing period
    sorted_regular = sorted(period_blocks["T_regular"], key=lambda t: t[0])
    sorted_lines = [line for _, line in sorted_regular]

    # Write to file
    with open(filename, "w") as f:
        f.writelines(period_blocks["T_zero"])
        f.writelines(period_blocks["T_inf"])
        f.writelines(sorted_lines)


def _format_excitation_line(
    period: float, beta_deg: float, i_dof: int, force: complex
) -> str:
    """Format a WAMIT excitation line.

    Parameters
    ----------
    period: float
        Wave period.
    beta_deg: float
        Wave direction (degrees).
    i_dof: int
        Degree of freedom index.
    force: complex
        Excitation force (complex).

    Returns
    -------
    str
        Formatted excitation line.
    """
    force_conj = np.conj(force)
    mod_f = np.abs(force_conj)
    phi_f = np.degrees(np.angle(force_conj))
    return "{:12.6e}\t{:12.6f}\t{:5d}\t{:12.6e}\t{:12.3f}\t{:12.6e}\t{:12.6e}\n".format(
        period, beta_deg, i_dof, mod_f, phi_f, force_conj.real, force_conj.imag
    )


def _write_wamit_excitation_line(
    f: TextIO,
    freq_key: str,
    freq_val: float,
    period: float,
    beta: float,
    dof: str,
    field: xarray.DataArray,
    rho: float,
    g: float,
    wave_amplitude: float,
    length_scale: float,
):
    """Write a single line for WAMIT .3 file format, using freq_key (omega or period)."""
    beta_deg = np.degrees(beta)
    i_dof = DOF_INDEX.get(dof)
    if i_dof is None:
        raise KeyError(f"DOF '{dof}' is not recognized in DOF_INDEX mapping.")

    dof_type = DOF_TYPE.get(dof, "trans")
    m = 2 if dof_type == "trans" else 3
    norm = rho * g * wave_amplitude * (length_scale**m)

    # Select value using appropriate key (omega or period)
    force = field.sel(
        {freq_key: freq_val, "wave_direction": beta, "influenced_dof": dof}
    ).item()
    force_normalized = force / norm

    line = _format_excitation_line(period, beta_deg, i_dof, force_normalized)
    f.write(line)


def _export_wamit_excitation_force(
    dataset: xarray.Dataset,
    field_name: str,
    filename: str,
    length_scale: float = 1.0,
    wave_amplitude: float = 1.0,
):
    """
    Export excitation-like forces to a WAMIT .3-style file.

    Format:
        PER     BETA    I     Fmagnitude  Fphase  Freal  Fimaginary
    """
    forward_speed = dataset["forward_speed"].values
    if not np.isclose(forward_speed, 0.0):
        raise ValueError("Forward speed must be zero for WAMIT export.")

    if field_name not in dataset:
        raise ValueError(f"Missing field '{field_name}' in dataset.")

    field = dataset[field_name]
    rho = dataset["rho"].item()
    betas = field.coords["wave_direction"].values
    dofs = list(field.coords["influenced_dof"].values)
    g = dataset["g"].item()

    # Determine main frequency coordinate
    freq_key, freq_vals, period_vals = identify_frequency_axis(dataset=dataset)

    # Sort by increasing period
    sorted_indices = np.argsort(period_vals)
    sorted_periods = period_vals[sorted_indices]
    sorted_freqs = freq_vals[sorted_indices]

    with open(filename, "w") as f:
        for freq_val, period in zip(sorted_freqs, sorted_periods):
            # Skip WAMIT special cases
            if np.isclose(freq_val, 0.0) or np.isinf(freq_val):
                continue
            for beta in betas:
                for dof in dofs:
                    _write_wamit_excitation_line(
                        f=f,
                        freq_key=freq_key,
                        freq_val=freq_val,
                        period=period,
                        beta=beta,
                        dof=dof,
                        field=field,
                        rho=rho,
                        g=g,
                        wave_amplitude=wave_amplitude,
                        length_scale=length_scale,
                    )


def export_wamit_3(dataset: xarray.Dataset, filename: str) -> None:
    """Export total excitation to WAMIT .3 file.

    Parameters
    ----------
    dataset: xarray.Dataset
        Dataset containing the desired complex-valued force field.
    filename: str
        Output path for the .3 file.
    """
    _export_wamit_excitation_force(dataset, "excitation_force", filename)


def export_wamit_3fk(dataset: xarray.Dataset, filename: str) -> None:
    """Export Froude-Krylov contribution to WAMIT .3fk file.

    Parameters
    ----------
    dataset: xarray.Dataset
        Dataset containing the desired complex-valued force field.
    filename: str
        Output path for the .3fk file.
    """
    _export_wamit_excitation_force(dataset, "Froude_Krylov_force", filename)


def export_wamit_3sc(dataset: xarray.Dataset, filename: str) -> None:
    """Export scattered (diffraction) contribution to WAMIT .3sc file.

    Parameters
    ----------
    dataset: xarray.Dataset
        Dataset containing the desired complex-valued force field.
    filename: str
        Output path for the .3sc file.
    """
    _export_wamit_excitation_force(dataset, "diffraction_force", filename)


def export_to_wamit(
    dataset: xarray.Dataset,
    problem_name: str,
    exports: Iterable[str] = ("1", "3", "3fk", "3sc", "hst"),
) -> None:
    """
    Master function to export a Capytaine dataset to WAMIT-format files.

    Parameters
    ----------
    dataset: xarray.Dataset
        Dataset containing the desired complex-valued force field.
    problem_name: str
        Base filename for WAMIT files (e.g. "output" → output.1, output.3fk, etc.).
    exports: iterable of str
        Which files to export: any combination of "1", "3", "3fk", "3sc", "hst".
    """
    export_map = {
        "1": ("radiation coefficients", export_wamit_1, ".1"),
        "3": ("total excitation force", export_wamit_3, ".3"),
        "3fk": ("Froude-Krylov force", export_wamit_3fk, ".3fk"),
        "3sc": ("diffraction force", export_wamit_3sc, ".3sc"),
        "hst": ("hydrostatics", export_wamit_hst, ".hst"),
    }
    check_dataset_ready_for_export(dataset)

    for key in exports:
        if key not in export_map:
            logger.warning(
                f"Export to WAMIT format: unknown option '{key}' — skipping."
            )
            continue

        description, func, ext = export_map[key]
        filepath = f"{problem_name}{ext}"

        try:
            func(dataset, filepath)
            logger.info(f"Export to WAMIT format: exported {filepath} ({description})")
        except Exception as e:
            logger.warning(f"Export to WAMIT format: did not export {filepath}: {e}")



import pytest

import numpy as np
import capytaine as cpt


def test_irr_freq_warning_no_lid():
    mesh = cpt.mesh_parallelepiped(size=(1, 1, 1)).immersed_part()
    body = cpt.FloatingBody(mesh=mesh, lid_mesh=None, dofs=cpt.rigid_body_dofs())
    assert 6.0 < body.first_irregular_frequency_estimate() < 7.0


def test_irr_freq_warning_subsurface_lid():
    mesh = cpt.mesh_parallelepiped(size=(1, 1, 1)).immersed_part()
    lid_mesh = cpt.mesh_rectangle(size=(1, 1), center=(0, 0, -0.1))
    body = cpt.FloatingBody(mesh=mesh, lid_mesh=lid_mesh, dofs=cpt.rigid_body_dofs())
    assert 10.0 < body.first_irregular_frequency_estimate() < 11.0


def test_irr_freq_warning_surface_lid():
    mesh = cpt.mesh_parallelepiped(size=(1, 1, 1)).immersed_part()
    lid_mesh = cpt.mesh_rectangle(size=(1, 1), center=(0, 0, 0))
    body = cpt.FloatingBody(mesh=mesh, lid_mesh=lid_mesh, dofs=cpt.rigid_body_dofs())
    assert body.first_irregular_frequency_estimate() == np.inf


def test_lid_with_upwards_normals():
    mesh = cpt.mesh_parallelepiped(center=(0.0, 0.0, -0.4)).immersed_part()
    lid_mesh = cpt.mesh_rectangle(size=(1, 1), resolution=(4, 4), center=(0, 0, -0.1), normal=(0, 0, 1))
    # Like body_with_lid() but with upward normals
    body = cpt.FloatingBody(mesh, lid_mesh=lid_mesh, dofs=cpt.rigid_body_dofs())
    np.testing.assert_allclose(body.lid_mesh.faces_normals[:, 2], -np.ones((body.lid_mesh.nb_faces,)))

    # Old lid_mesh is unchanged
    assert body.lid_mesh is not lid_mesh
    np.testing.assert_allclose(lid_mesh.faces_normals[:, 2], np.ones((body.lid_mesh.nb_faces,)))


@pytest.fixture
def body_without_lid():
    mesh = cpt.mesh_parallelepiped(center=(0.0, 0.0, -0.4)).immersed_part()
    body_without_lid = cpt.FloatingBody(mesh, dofs=cpt.rigid_body_dofs())
    return body_without_lid


@pytest.fixture
def body_with_lid():
    mesh = cpt.mesh_parallelepiped(center=(0.0, 0.0, -0.4)).immersed_part()
    lid_mesh = cpt.mesh_rectangle(size=(1, 1), resolution=(4, 4), center=(0, 0, -0.1), normal=(0, 0, -1))
    body_with_lid = cpt.FloatingBody(mesh, lid_mesh=lid_mesh, dofs=cpt.rigid_body_dofs())
    return body_with_lid


def test_effect_of_lid_on_froude_krylov_force(body_without_lid, body_with_lid):
    # Froude-Krylov force should be unchanged by the lid
    from capytaine.bem.airy_waves import froude_krylov_force
    pb_without_lid = cpt.DiffractionProblem(body=body_without_lid)
    pb_with_lid = cpt.DiffractionProblem(body=body_with_lid)

    fk_with_lid = froude_krylov_force(pb_with_lid)["Heave"]
    fk_without_lid = froude_krylov_force(pb_without_lid)["Heave"]
    assert fk_with_lid == pytest.approx(fk_without_lid)


def test_effect_of_lid_on_matrices(body_without_lid, body_with_lid):
    n_hull_mesh = body_without_lid.mesh.nb_faces
    solver = cpt.BEMSolver(green_function=cpt.Delhommeau(gf_singularities='low_freq'))

    params = [0.0, np.inf, 1.0, solver.green_function]
    S_with, K_with = solver.engine.build_matrices(
            body_with_lid.mesh_including_lid, body_with_lid.mesh_including_lid,
            *params, adjoint_double_layer=True,
            )
    S_without, K_without = solver.engine.build_matrices(
            body_without_lid.mesh, body_without_lid.mesh,
            *params, adjoint_double_layer=True,
            )
    _, D_with = solver.engine.build_matrices(
            body_with_lid.mesh_including_lid, body_with_lid.mesh_including_lid,
            *params, adjoint_double_layer=False,
            )
    _, D_without = solver.engine.build_matrices(
            body_without_lid.mesh, body_without_lid.mesh,
            *params, adjoint_double_layer=False,
            )

    np.testing.assert_allclose(K_with[:n_hull_mesh, :n_hull_mesh], K_without, atol=1e-8)
    np.testing.assert_allclose(D_with[:n_hull_mesh, :n_hull_mesh], D_without, atol=1e-8)
    np.testing.assert_allclose(S_with[:n_hull_mesh, :n_hull_mesh], S_without, atol=1e-8)


@pytest.mark.parametrize("water_depth", [np.inf, 10.0])
@pytest.mark.parametrize("forward_speed", [0.0, 1.0])
def test_effect_of_lid_on_regular_frequency_diffraction_force(
        body_without_lid, body_with_lid, water_depth, forward_speed,
        ):
    solver = cpt.BEMSolver(green_function=cpt.Delhommeau(gf_singularities='low_freq'))

    pb_with = cpt.DiffractionProblem(
            body=body_with_lid, wavelength=3.0,
            water_depth=water_depth, forward_speed=forward_speed,
            )
    res_with = solver.solve(pb_with)
    f_with = res_with.forces["Heave"]

    pb_without = cpt.DiffractionProblem(
            body=body_without_lid, wavelength=3.0,
            water_depth=water_depth, forward_speed=forward_speed
            )
    res_without = solver.solve(pb_without)
    f_without = res_without.forces["Heave"]

    assert f_with == pytest.approx(f_without, rel=5e-2)


def test_effect_of_lid_on_infinite_frequency(
        body_without_lid, body_with_lid,
        ):
    solver = cpt.BEMSolver(green_function=cpt.Delhommeau(gf_singularities='low_freq'))

    pb_with = cpt.RadiationProblem(
            body=body_with_lid, wavenumber=np.inf, radiating_dof="Heave"
            )
    res_with = solver.solve(pb_with)
    f_with = res_with.forces["Heave"]

    pb_without = cpt.RadiationProblem(
            body=body_without_lid, wavenumber=np.inf, radiating_dof="Heave"
            )
    res_without = solver.solve(pb_without)
    f_without = res_without.forces["Heave"]

    assert f_with == pytest.approx(f_without, rel=5e-2)


@pytest.mark.parametrize("water_depth", [np.inf, 10.0])
@pytest.mark.parametrize("forward_speed", [0.0, 1.0])
def test_effect_of_lid_on_regular_frequency_free_surface_elevation(
        body_without_lid, body_with_lid,
        water_depth, forward_speed,
        ):
    solver = cpt.BEMSolver(green_function=cpt.Delhommeau(gf_singularities='low_freq'))
    point = np.array([[4.0, 4.0]])

    pb_with = cpt.DiffractionProblem(
            body=body_with_lid, wavelength=3.0,
            water_depth=water_depth, forward_speed=forward_speed,
            )
    res_with = solver.solve(pb_with)
    fse_with = solver.compute_free_surface_elevation(point, res_with)

    pb_without = cpt.DiffractionProblem(
            body=body_without_lid, wavelength=3.0,
            water_depth=water_depth, forward_speed=forward_speed,
            )
    res_without = solver.solve(pb_without)
    fse_without = solver.compute_free_surface_elevation(point, res_without)

    assert fse_with == pytest.approx(fse_without, rel=5e-2)


@pytest.mark.parametrize("water_depth", [np.inf, 10.0])
@pytest.mark.parametrize("forward_speed", [0.0, 1.0])
def test_effect_of_lid_on_regular_frequency_field_velocity(
        body_without_lid, body_with_lid,
        water_depth, forward_speed,
        ):
    solver = cpt.BEMSolver(green_function=cpt.Delhommeau(gf_singularities='low_freq'))
    point = np.array([[4.0, 4.0, -2.0]])

    pb_with = cpt.DiffractionProblem(
            body=body_with_lid, wavelength=3.0,
            water_depth=water_depth, forward_speed=forward_speed,
            )
    res_with = solver.solve(pb_with)
    u_with = solver.compute_velocity(point, res_with)

    pb_without = cpt.DiffractionProblem(
            body=body_without_lid, wavelength=3.0,
            water_depth=water_depth, forward_speed=forward_speed,
            )
    res_without = solver.solve(pb_without)
    u_without = solver.compute_velocity(point, res_without)

    assert u_with == pytest.approx(u_without, rel=5e-2)


def test_lid_multibody(body_with_lid):
    two_bodies = body_with_lid + body_with_lid.translated_x(5.0)
    assert isinstance(two_bodies.mesh, cpt.CollectionOfMeshes)
    assert isinstance(two_bodies.lid_mesh, cpt.CollectionOfMeshes)

    # Check that the lid are still positioned on top of each component
    def horizontal_center(mesh):
        return mesh.faces_centers.mean(axis=0)[:2]
    np.testing.assert_allclose(
        horizontal_center(two_bodies.lid_mesh[0]),
        horizontal_center(two_bodies.mesh[0]),
        atol=1e-6
    )
    np.testing.assert_allclose(
        horizontal_center(two_bodies.lid_mesh[1]),
        horizontal_center(two_bodies.mesh[1]),
        atol=1e-6
    )

    pb = cpt.DiffractionProblem(body=two_bodies, wavelength=3.0)
    solver = cpt.BEMSolver(green_function=cpt.Delhommeau(gf_singularities='low_freq'))
    solver.solve(pb)


def test_lid_with_plane_symmetry():
    mesh = cpt.mesh_horizontal_cylinder(reflection_symmetry=True).immersed_part()
    lid_mesh = cpt.ReflectionSymmetricMesh(
            cpt.mesh_rectangle(size=(1.0, 10,), faces_max_radius=0.5, center=(0, 0.5, -0.05,)),
            plane=mesh.plane
            )
    body = cpt.FloatingBody(mesh=mesh, lid_mesh=lid_mesh, dofs=cpt.rigid_body_dofs())
    pb = cpt.RadiationProblem(body=body, wavelength=1.0, radiating_dof="Heave")
    solver = cpt.BEMSolver()
    S, K = solver.engine.build_matrices(pb.body.mesh_including_lid, pb.body.mesh_including_lid,
                                        pb.free_surface, pb.water_depth, pb.wavenumber,
                                        solver.green_function)
    from capytaine.matrices.block_toeplitz import BlockSymmetricToeplitzMatrix
    assert isinstance(S, BlockSymmetricToeplitzMatrix)
    assert isinstance(K, BlockSymmetricToeplitzMatrix)

    res = solver.solve(pb)

    ref_body = cpt.FloatingBody(mesh=mesh.merged(), lid_mesh=lid_mesh.merged(), dofs=cpt.rigid_body_dofs())
    ref_pb = cpt.RadiationProblem(body=ref_body, wavelength=1.0, radiating_dof="Heave")
    ref_res = solver.solve(ref_pb)
    assert res.force["Heave"] == pytest.approx(ref_res.force["Heave"])


def test_lid_with_nested_plane_symmetry():
    mesh = cpt.mesh_parallelepiped(center=(0, 0, -1.0), reflection_symmetry=True)
    hull_mesh, lid_mesh = mesh.extract_lid()
    body = cpt.FloatingBody(mesh=mesh, lid_mesh=lid_mesh, dofs=cpt.rigid_body_dofs())
    pb = cpt.RadiationProblem(body=body, wavelength=1.0)
    solver = cpt.BEMSolver()
    S, K = solver.engine.build_matrices(pb.body.mesh_including_lid, pb.body.mesh_including_lid,
                                        pb.free_surface, pb.water_depth, pb.wavenumber,
                                        solver.green_function)
    from capytaine.matrices.block_toeplitz import BlockSymmetricToeplitzMatrix
    assert isinstance(S, BlockSymmetricToeplitzMatrix)
    assert isinstance(S.all_blocks[0, 0], BlockSymmetricToeplitzMatrix)
    assert isinstance(K, BlockSymmetricToeplitzMatrix)
    assert isinstance(K.all_blocks[0, 0], BlockSymmetricToeplitzMatrix)


@pytest.mark.parametrize("water_depth", [np.inf, 10.0])
def test_panel_on_free_surface(water_depth):
    mesh = cpt.mesh_parallelepiped(center=(0.0, 0.0, -0.5))
    body = cpt.FloatingBody(mesh, cpt.rigid_body_dofs())
    body.mesh.compute_quadrature("Gauss-Legendre 2")
    pb = cpt.RadiationProblem(body=body, wavelength=1.0, water_depth=water_depth, radiating_dof="Heave")
    cpt.BEMSolver(green_function=cpt.Delhommeau(gf_singularities="low_freq")).solve(pb)


def test_panel_on_free_surface_with_high_freq(caplog):
    mesh = cpt.mesh_rectangle(center=(0.0, 0.0, 0.0), size=(1.0, 1.0))
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs())
    pb = cpt.RadiationProblem(body=body, wavelength=1.0, radiating_dof="Heave")
    cpt.BEMSolver(green_function=cpt.Delhommeau(gf_singularities="low_freq")).solve(pb)
    with caplog.at_level("WARNING"):
        cpt.BEMSolver(green_function=cpt.Delhommeau(gf_singularities="high_freq")).solve(pb)
    assert "free surface panel" in caplog.text



""" This module contains a class to describe the 2D mesh of the surface of a body in a 3D space.
Based on meshmagick <https://github.com/LHEEA/meshmagick> by François Rongère.
"""
# Copyright (C) 2017-2019 Matthieu Ancellin, based on the work of François Rongère
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging
from itertools import count

import numpy as np

from capytaine.meshes.geometry import Abstract3DObject, ClippableMixin, Plane, inplace_transformation, xOy_Plane
from capytaine.meshes.properties import compute_faces_properties, connected_components, connected_components_of_waterline
from capytaine.meshes.surface_integrals import SurfaceIntegralsMixin
from capytaine.meshes.quality import (merge_duplicates, heal_normals, remove_unused_vertices,
                                      heal_triangles, remove_degenerated_faces)
from capytaine.tools.optional_imports import import_optional_dependency
from capytaine.meshes.quadratures import compute_quadrature_on_faces

LOG = logging.getLogger(__name__)


class Mesh(ClippableMixin, SurfaceIntegralsMixin, Abstract3DObject):
    """A class to handle unstructured 2D meshes in a 3D space.

    Parameters
    ----------
    vertices : array_like of shape (nv, 3)
        Array of mesh vertices coordinates.Each line of the array represents one vertex
        coordinates
    faces : array_like of shape (nf, 4)
        Arrays of mesh connectivities for faces. Each line of the array represents indices of
        vertices that form the face, expressed in counterclockwise order to ensure outward normals
        description.
    name : str, optional
        The name of the mesh. If None, the mesh is given an automatic name based on its internal ID.
    quadrature_method: None or str or Quadpy quadrature, optional
        The method used to compute quadrature points in each cells.
        By default: None, that is a one-point first order scheme is used.
    """

    _ids = count(0)  # A counter for automatic naming of new meshes.

    def __init__(self, vertices=None, faces=None, name=None, *, quadrature_method=None):

        if vertices is None or len(vertices) == 0:
            vertices = np.zeros((0, 3))

        if faces is None or len(faces) == 0:
            faces = np.zeros((0, 4))

        if name is None:
            self.name = f'mesh_{next(Mesh._ids)}'
        else:
            self.name = str(name)

        self.__internals__ = dict()
        self.vertices = vertices  # Not a direct assignment, goes through the setter method below.
        self.faces = faces  # Not a direct assignment, goes through the setter method below.

        LOG.debug(f"New mesh: {repr(self)}")

        self.quadrature_method = quadrature_method

    def __short_str__(self):
        return (f"{self.__class__.__name__}(..., name=\"{self.name}\")")

    def __str__(self):
        return (f"{self.__class__.__name__}(vertices=[[... {self.nb_vertices} vertices ...]], "
                f"faces=[[... {self.nb_faces} faces ...]], name=\"{self.name}\")")

    def __repr__(self):
        # shift = len(self.__class__.__name__) + 1
        # vert_str = np.array_repr(self.vertices).replace('\n', '\n' + (shift + 9)*' ')
        # faces_str = np.array_repr(self.faces).replace('\n', '\n' + (shift + 6)*' ')
        # return f"{self.__class__.__name__}(\n{' '*shift}vertices={vert_str},\n{' '*shift}faces={faces_str}\n{' '*shift}name=\"{self.name}\"\n)"
        return (f"{self.__class__.__name__}(vertices=[[... {self.nb_vertices} vertices ...]], "
                f"faces=[[... {self.nb_faces} faces ...]], name=\"{self.name}\")")

    def _repr_pretty_(self, p, cycle):
        p.text(self.__str__())

    def __rich_repr__(self):
        class CustomRepr:
            def __init__(self, n, kind):
                self.n = n
                self.kind = kind
            def __repr__(self):
                return "[[... {} {} ...]]".format(self.n, self.kind)
        yield "vertices", CustomRepr(self.nb_vertices, "vertices")
        yield "faces", CustomRepr(self.nb_faces, "faces")
        yield "name", self.name

    @property
    def nb_vertices(self) -> int:
        """Get the number of vertices in the mesh."""
        return self._vertices.shape[0]

    @property
    def vertices(self) -> np.ndarray:
        """Get the vertices array coordinate of the mesh."""
        return self._vertices

    @vertices.setter
    def vertices(self, value) -> None:
        self._vertices = np.array(value, dtype=float)
        assert self._vertices.shape[1] == 3, \
            "Vertices of a mesh should be provided as a sequence of 3-ple."
        self.__internals__.clear()

    @property
    def nb_faces(self) -> int:
        """Get the number of faces in the mesh."""
        return self._faces.shape[0]

    @property
    def faces(self) -> np.ndarray:
        """Get the faces connectivity array of the mesh."""
        return self._faces

    @faces.setter
    def faces(self, faces):
        faces = np.array(faces, dtype=int)
        assert np.all(faces >= 0), \
            "Faces of a mesh should be provided as positive integers (ids of vertices)"
        assert faces.shape[1] == 4, \
            "Faces of a mesh should be provided as a sequence of 4-ple."
        assert len(faces) == 0 or faces.max()+1 <= self.nb_vertices, \
            "The array of faces should only reference vertices that are in the mesh."
        self._faces = faces
        self.__internals__.clear()

    def copy(self, name=None) -> 'Mesh':
        """Get a copy of the current mesh instance.

        Parameters
        ----------
        name : string, optional
            a name for the new mesh

        Returns
        -------
        Mesh
            mesh instance copy
        """
        from copy import deepcopy
        new_mesh = deepcopy(self)
        if name is not None:
            new_mesh.name = name
        return new_mesh

    def merged(self):
        """Dummy method to be generalized for collections of meshes."""
        return self

    def tree_view(self, **kwargs):
        """Dummy method to be generalized for collections of meshes."""
        return self.__short_str__()

    def path_to_leaf(self):
        """Dummy method to be generalized for collection of meshes."""
        return [[]]

    def to_meshmagick(self):
        """Convert the Mesh object as a Mesh object from meshmagick.
        Mostly for debugging."""
        from meshmagick.mesh import Mesh
        meshmagick_mesh = Mesh(self.vertices, self.faces, name=self.name)
        meshmagick_mesh.heal_mesh()
        return meshmagick_mesh

    ##################
    #  Extract face  #
    ##################

    def get_face(self, face_id):
        """Get the face described by its vertices connectivity.

        Parameters
        ----------
        face_id : int
            Face id

        Returns
        -------
        ndarray
            If the face is a triangle, the array has 3 components, else it has 4 (quadrangle)
        """
        if self.is_triangle(face_id):
            return self._faces[face_id, :3]
        else:
            return self._faces[face_id]

    def extract_one_face(self, id_face):
        vertices = self.vertices[self.faces[id_face, :], :]
        mesh = Mesh(vertices=vertices, faces=np.array([[0, 1, 2, 3]]), name=f"single_face_from_{self.name}")

        for prop in self.__internals__:
            if prop[:4] == "face":
                mesh.__internals__[prop] = self.__internals__[prop][[id_face]]

        return mesh

    def extract_faces(self, id_faces_to_extract, return_index=False, name=None):
        """
        Extracts a new mesh from a selection of faces ids

        Parameters
        ----------
        id_faces_to_extract : ndarray
            Indices of faces that have to be extracted
        return_index: bool, optional
            Flag to output old indices
        name: string, optional
            Name for the new mesh

        Returns
        -------
        Mesh
            A new Mesh instance composed of the extracted faces
        """
        nv = self.nb_vertices

        # Determination of the vertices to keep
        vertices_mask = np.zeros(nv, dtype=bool)
        vertices_mask[self._faces[id_faces_to_extract].flatten()] = True
        id_v = np.arange(nv)[vertices_mask]

        # Building up the vertex array
        v_extracted = self._vertices[id_v]
        new_id__v = np.arange(nv)
        new_id__v[id_v] = np.arange(len(id_v))

        faces_extracted = self._faces[id_faces_to_extract]
        faces_extracted = new_id__v[faces_extracted.flatten()].reshape((len(id_faces_to_extract), 4))

        extracted_mesh = Mesh(v_extracted, faces_extracted)

        for prop in self.__internals__:
            if prop[:4] == "face":
                extracted_mesh.__internals__[prop] = self.__internals__[prop][id_faces_to_extract]

        if name is None:
            if self.name is not None and self.name.startswith("mesh_extracted_from_"):
                extracted_mesh.name = self.name
            else:
                extracted_mesh.name = f"mesh_extracted_from_{self.name}"
        else:
            extracted_mesh.name = name

        if return_index:
            return extracted_mesh, id_v
        else:
            return extracted_mesh

    def sliced_by_plane(self, plane: Plane):
        from capytaine.meshes.collections import CollectionOfMeshes
        faces_ids_on_one_side = np.where(plane.distance_to_point(self.faces_centers) < 0)[0]
        if len(faces_ids_on_one_side) == 0 or len(faces_ids_on_one_side) == self.nb_faces:
            return self.copy()
        else:
            mesh_part_1 = self.extract_faces(faces_ids_on_one_side)
            mesh_part_2 = self.extract_faces(list(set(range(self.nb_faces)) - set(faces_ids_on_one_side)))
            return CollectionOfMeshes([mesh_part_1, mesh_part_2],
                                      name=f"{self.name}_splitted_by_{plane}")


    #####################
    #  Mean and radius  #
    #####################

    @property
    def center_of_mass_of_nodes(self):
        """(Non-weighted) center of mass of the nodes of the mesh."""
        if 'center_of_mass_of_nodes' not in self.__internals__:
            center_of_mass_of_nodes = np.mean(self.vertices, axis=0)
            self.__internals__['center_of_mass_of_nodes'] = center_of_mass_of_nodes
            return center_of_mass_of_nodes
        return self.__internals__['center_of_mass_of_nodes']

    @property
    def diameter_of_nodes(self):
        """Maximum distance between two nodes of the mesh."""
        if 'diameter_of_nodes' not in self.__internals__:
            diameter_of_nodes = 2*np.max(
                np.linalg.norm(self.vertices - self.center_of_mass_of_nodes, axis=-1)
            )
            self.__internals__['diameter_of_nodes'] = diameter_of_nodes
            return diameter_of_nodes
        return self.__internals__['diameter_of_nodes']

    ######################
    #  Faces properties  #
    ######################

    @property
    def faces_areas(self) -> np.ndarray:
        """Get the array of faces areas of the mesh."""
        if 'faces_areas' not in self.__internals__:
            self.__internals__.update(compute_faces_properties(self))
        return self.__internals__['faces_areas']

    @property
    def faces_centers(self) -> np.ndarray:
        """Get the array of faces centers of the mesh."""
        if 'faces_centers' not in self.__internals__:
            self.__internals__.update(compute_faces_properties(self))
        return self.__internals__['faces_centers']

    @property
    def faces_normals(self) -> np.ndarray:
        """Get the array of faces normals of the mesh."""
        if 'faces_normals' not in self.__internals__:
            self.__internals__.update(compute_faces_properties(self))
        return self.__internals__['faces_normals']

    @property
    def faces_radiuses(self) -> np.ndarray:
        """Get the array of faces radiuses of the mesh."""
        if 'faces_radiuses' not in self.__internals__:
            self.__internals__.update(compute_faces_properties(self))
        return self.__internals__['faces_radiuses']

    @property
    def quadrature_points(self):
        if 'quadrature' not in self.__internals__:
            self.compute_quadrature(self.quadrature_method)
        return self.__internals__['quadrature']

    def compute_quadrature(self, method):
        self.heal_triangles()
        all_faces = self.vertices[self.faces[:, :], :]
        if method is None:
            points = self.faces_centers.reshape((self.nb_faces, 1, 3))
            weights = self.faces_areas.reshape((self.nb_faces, 1))
        else:
            points, weights = compute_quadrature_on_faces(all_faces, method)
        self.__internals__['quadrature'] = (points, weights)
        self.quadrature_method = method
        return points, weights


    ###############################
    #  Triangles and quadrangles  #
    ###############################

    def is_triangle(self, face_id) -> bool:
        """Returns if a face is a triangle

        Parameters
        ----------
        face_id : int
            Face id
        """
        assert 0 <= face_id < self.nb_faces
        return self._faces[face_id, 0] == self._faces[face_id, -1]

    def _compute_triangles_quadrangles(self):
        triangle_mask = (self._faces[:, 0] == self._faces[:, -1])
        quadrangles_mask = np.invert(triangle_mask)
        triangles_quadrangles = {'triangles_ids': np.where(triangle_mask)[0],
                                 'quadrangles_ids': np.where(quadrangles_mask)[0]}
        self.__internals__.update(triangles_quadrangles)

    @property
    def triangles_ids(self) -> np.ndarray:
        """Get the array of ids of triangle shaped faces."""
        if 'triangles_ids' not in self.__internals__:
            self._compute_triangles_quadrangles()
        return self.__internals__['triangles_ids']

    @property
    def nb_triangles(self) -> int:
        """Get the number of triangles in the mesh."""
        if 'triangles_ids'not in self.__internals__:
            self._compute_triangles_quadrangles()
        return len(self.__internals__['triangles_ids'])

    @property
    def quadrangles_ids(self) -> np.ndarray:
        """Get the array of ids of quadrangle shaped faces."""
        if 'triangles_ids' not in self.__internals__:
            self._compute_triangles_quadrangles()
        return self.__internals__['quadrangles_ids']

    @property
    def nb_quadrangles(self) -> int:
        """Get the number of quadrangles in the mesh."""
        if 'triangles_ids' not in self.__internals__:
            self._compute_triangles_quadrangles()
        return len(self.__internals__['quadrangles_ids'])

    #############
    #  Display  #
    #############

    @property
    def axis_aligned_bbox(self):
        """Get the axis aligned bounding box of the mesh.

        Returns
        -------
        tuple
            (xmin, xmax, ymin, ymax, zmin, zmax)
        """
        if self.nb_vertices > 0:
            x, y, z = self._vertices.T
            return (x.min(), x.max(),
                    y.min(), y.max(),
                    z.min(), z.max())
        else:
            return tuple(np.zeros(6))

    @property
    def squared_axis_aligned_bbox(self):
        """Get a squared axis aligned bounding box of the mesh.

        Returns
        -------
        tuple
            (xmin, xmax, ymin, ymax, zmin, zmax)

        Note
        ----
        This method differs from `axis_aligned_bbox()` by the fact that
        the bounding box that is returned is squared but have the same center as the `axis_aligned_bbox()`.
        """
        xmin, xmax, ymin, ymax, zmin, zmax = self.axis_aligned_bbox
        (x0, y0, z0) = np.array([xmin+xmax, ymin+ymax, zmin+zmax]) * 0.5
        d = (np.array([xmax-xmin, ymax-ymin, zmax-zmin]) * 0.5).max()

        return x0-d, x0+d, y0-d, y0+d, z0-d, z0+d

    def show(self, **kwargs):
        self.show_vtk(**kwargs)

    def show_vtk(self, **kwargs):
        """Shows the mesh in the vtk viewer"""
        from capytaine.ui.vtk.mesh_viewer import MeshViewer

        viewer = MeshViewer()
        viewer.add_mesh(self, **kwargs)
        viewer.show()
        viewer.finalize()

    def show_matplotlib(self, ax=None,
                        normal_vectors=False, scale_normal_vector=None,
                        saveas=None, color_field=None, cmap=None,
                        cbar_label=None,
                        **kwargs):
        """Poor man's viewer with matplotlib.

        Parameters
        ----------
        ax: matplotlib axis
            The 3d axis in which to plot the mesh. If not provided, create a new one.
        normal_vectors: bool
            If True, print normal vector.
        scale_normal_vector: array of shape (nb_faces, )
            Scale separately each of the normal vectors.
        saveas: str
            File path where to save the image.
        color_field: array of shape (nb_faces, )
            Scalar field to be plot on the mesh (optional).
        cmap: matplotlib colormap
            Colormap to use for field plotting.
        cbar_label: string
            Label for colormap

        Other parameters are passed to Poly3DCollection.
        """
        matplotlib = import_optional_dependency("matplotlib")
        import importlib
        plt = importlib.import_module("matplotlib.pyplot")
        cm = importlib.import_module("matplotlib.cm")

        mpl_toolkits = import_optional_dependency("mpl_toolkits", package_name="matplotlib")
        Poly3DCollection = mpl_toolkits.mplot3d.art3d.Poly3DCollection

        default_axis = ax is None
        if default_axis:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection="3d")

        faces = []
        for face in self.faces:
            vertices = []
            for index_vertex in face:
                vertices.append(self.vertices[int(index_vertex), :])
            faces.append(vertices)

        if color_field is None:
            if 'facecolors' not in kwargs:
                kwargs['facecolors'] = "yellow"
        else:
            if cmap is None:
                cmap = matplotlib.colormaps['coolwarm']
            m = cm.ScalarMappable(cmap=cmap)
            m.set_array([min(color_field), max(color_field)])
            m.set_clim(vmin=min(color_field), vmax=max(color_field))
            colors = m.to_rgba(color_field)
            kwargs['facecolors'] = colors
        if 'edgecolor' not in kwargs:
            kwargs['edgecolor'] = 'k'

        ax.add_collection3d(Poly3DCollection(faces, **kwargs))

        if color_field is not None:
            cbar = plt.colorbar(m, ax=ax)
            if cbar_label is not None:
                cbar.set_label(cbar_label)



        # Plot normal vectors.
        if normal_vectors:
            if scale_normal_vector is not None:
                vectors = (scale_normal_vector * self.faces_normals.T).T
            else:
                vectors = self.faces_normals
            ax.quiver(*zip(*self.faces_centers), *zip(*vectors), length=0.2)


        ax.set_xlabel("x")
        ax.set_ylabel("y")
        ax.set_zlabel("z")

        xmin, xmax, ymin, ymax, zmin, zmax = self.squared_axis_aligned_bbox
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)
        ax.set_zlim(zmin, zmax)

        if default_axis:
            if saveas is not None:
                plt.tight_layout()
                plt.savefig(saveas)
            else:
                plt.show()

    ################################
    #  Transformation of the mesh  #
    ################################

    @inplace_transformation
    def translate(self, vector) -> 'Mesh':
        """Translates the mesh in 3D giving the 3 distances along coordinate axes.

        Parameters
        ----------
        vector : array_like
            translation vector
        """
        vector = np.asarray(vector, dtype=float)
        assert vector.shape == (3,), "The translation vector should be given as a 3-ple of values."

        self.vertices += vector

        return self

    @inplace_transformation
    def rotate(self, axis, angle) -> 'Mesh':
        """Rotate the mesh of a given angle around an axis.

        Parameters
        ----------
        axis : Axis
        angle : float
        """

        self._vertices = axis.rotate_points(self._vertices, angle)

        return self

    # OTHER
    @inplace_transformation
    def flip_normals(self) -> 'Mesh':
        """Flips every normals of the mesh."""

        self._faces = np.fliplr(self._faces)

        return self

    @inplace_transformation
    def mirror(self, plane) -> 'Mesh':
        """Flip the mesh with respect to a plane.

        Parameters
        ----------
        plane : Plane
            The mirroring plane
        """
        self.vertices -= 2 * np.outer(np.dot(self.vertices, plane.normal) - plane.c, plane.normal)
        self.flip_normals()
        return self

    def symmetrized(self, plane):
        from capytaine.meshes.symmetric import ReflectionSymmetricMesh
        half = self.clipped(plane, name=f"{self.name}_half")
        return ReflectionSymmetricMesh(half, plane=plane, name=f"symmetrized_of_{self.name}")

    @inplace_transformation
    def clip(self, plane) -> 'Mesh':
        from capytaine.meshes.clipper import clip
        clipped_self = clip(self, plane=plane)
        self.vertices = clipped_self.vertices
        self.faces = clipped_self.faces
        self._clipping_data = clipped_self._clipping_data
        return self

    @inplace_transformation
    def triangulate_quadrangles(self) -> 'Mesh':
        """Triangulates every quadrangles of the mesh by simple splitting.
        Each quadrangle gives two triangles.

        Note
        ----
        No checking on the triangle quality is done.
        """
        # Defining both triangles id lists to be generated from quadrangles
        t1 = (0, 1, 2)
        t2 = (0, 2, 3)

        faces = self._faces

        # Triangulation
        new_faces = faces[self.quadrangles_ids].copy()
        new_faces[:, :3] = new_faces[:, t1]
        new_faces[:, -1] = new_faces[:, 0]

        faces[self.quadrangles_ids, :3] = faces[:, t2][self.quadrangles_ids]
        faces[self.quadrangles_ids, -1] = faces[self.quadrangles_ids, 0]

        faces = np.concatenate((faces, new_faces))

        LOG.info('\nTriangulating quadrangles')
        if self.nb_quadrangles != 0:
            LOG.info('\t-->{:d} quadrangles have been split in triangles'.format(self.nb_quadrangles))

        self._faces = faces

        return self

    ####################
    #  Combine meshes  #
    ####################

    def join_meshes(*meshes, name=None, return_masks=False):
        from capytaine.meshes.collections import CollectionOfMeshes
        coll = CollectionOfMeshes(meshes, name=name)
        if return_masks:
            masks = []
            for i_mesh in range(len(meshes)):
                mask = np.full((coll.nb_faces,), False)
                mask[coll.indices_of_mesh(i_mesh)] = True
                masks.append(mask)
            return coll.merged(), masks
        else:
            return coll.merged()

    def __add__(self, mesh_to_add) -> 'Mesh':
        return self.join_meshes(mesh_to_add)

    ####################
    #  Compare meshes  #
    ####################
    # The objective is to write a mesh as a set of faces in order to check for equality or to
    # compute differences of meshes. Each face can be represented as a 4x3 array (4 triplets of
    # coordinates).
    # However, it is tricky on several aspects:
    #   * The builtin set class compares the hashes of its objects. Since numpy ndarray are not
    #   hashable, the 4x3 array can be transformed into a tuple of tuples (which is hashable).
    #   * Two faces with different numbering of the faces (but the same ordering) are recorded as
    #   different.
    #   * Two vertices equal up to machine precision can be recorded as different, due to rounding
    #   errors.
    #
    # A possible solution is to define a Face class and a Vertex class with the appropriate __eq__
    # and __hash__.
    #
    # The current implementation below is a rough draft.
    # However, the equality shall still be use for testing.

    def as_set_of_faces(self):
        return frozenset(frozenset(tuple(vertex) for vertex in face) for face in self.vertices[self.faces])

    @staticmethod
    def from_set_of_faces(set_of_faces):
        faces = []
        vertices = []
        for face in set_of_faces:
            ids_of_vertices_in_face = []

            for vertex in face:
                if vertex not in vertices:
                    i = len(vertices)
                    vertices.append(vertex)
                else:
                    i = vertices.index(vertex)
                ids_of_vertices_in_face.append(i)

            if len(ids_of_vertices_in_face) == 3:
                # Add a fourth node identical to the first one
                ids_of_vertices_in_face.append(ids_of_vertices_in_face[0])

            faces.append(ids_of_vertices_in_face)
        return Mesh(vertices=vertices, faces=faces)

    def __eq__(self, other):
        if not isinstance(other, Mesh):
            return NotImplemented
        else:
            return self.as_set_of_faces() == other.as_set_of_faces()

    def __hash__(self):
        if 'hash' not in self.__internals__:
            self.__internals__['hash'] = hash(self.as_set_of_faces())
        return self.__internals__['hash']

    ##################
    #  Mesh quality  #
    ##################

    def merge_duplicates(self, **kwargs):
        return merge_duplicates(self, **kwargs)

    def heal_normals(self, **kwargs):
        return heal_normals(self, **kwargs)

    def remove_unused_vertices(self, **kwargs):
        return remove_unused_vertices(self, **kwargs)

    def heal_triangles(self, **kwargs):
        return heal_triangles(self, **kwargs)

    def remove_degenerated_faces(self, **kwargs):
        return remove_degenerated_faces(self, **kwargs)

    @inplace_transformation
    def heal_mesh(self, closed_mesh=True):
        """Heals the mesh for different tests available.

        It applies:

        * Unused vertices removal
        * Degenerate faces removal
        * Duplicate vertices merging
        * Triangles healing
        * Normal healing
        """
        self.remove_unused_vertices()
        self.remove_degenerated_faces()
        self.merge_duplicates()
        self.heal_triangles()
        if closed_mesh:
            self.heal_normals()
        return self

    ##########
    #  Lids  #
    ##########

    def lowest_lid_position(self, omega_max, *, g=9.81):
        z_lid = 0.0
        for comp in connected_components(self):
            for ccomp in connected_components_of_waterline(comp):
                x_span = ccomp.vertices[:, 0].max() - ccomp.vertices[:, 0].min()
                y_span = ccomp.vertices[:, 1].max() - ccomp.vertices[:, 1].min()
                p = np.hypot(1/x_span, 1/y_span)
                z_lid_comp = -np.arctanh(np.pi*g*p/omega_max**2) / (np.pi * p)
                z_lid = min(z_lid, z_lid_comp)
        return 0.9*z_lid  # Add a small safety margin

    def generate_lid(self, z=0.0, faces_max_radius=None, name=None):
        """
        Return a mesh of the internal free surface of the body.

        Parameters
        ----------
        z: float, optional
            Vertical position of the lid. Default: 0.0
        faces_max_radius: float, optional
            resolution of the mesh of the lid.
            Default: mean of hull mesh resolution.
        name: str, optional
            A name for the new mesh

        Returns
        -------
        Mesh
            lid of internal surface
        """
        from capytaine.meshes.predefined.rectangles import mesh_rectangle

        clipped_hull_mesh = self.clipped(Plane(normal=(0, 0, 1), point=(0, 0, z)))
        # Alternatively: could keep only faces below z without proper clipping,
        # and it would work similarly.

        if clipped_hull_mesh.nb_faces == 0:
            return Mesh(None, None, name="lid for {}".format(self.name))

        x_span = clipped_hull_mesh.vertices[:, 0].max() - clipped_hull_mesh.vertices[:, 0].min()
        y_span = clipped_hull_mesh.vertices[:, 1].max() - clipped_hull_mesh.vertices[:, 1].min()
        x_mean = (clipped_hull_mesh.vertices[:, 0].max() + clipped_hull_mesh.vertices[:, 0].min())/2
        y_mean = (clipped_hull_mesh.vertices[:, 1].max() + clipped_hull_mesh.vertices[:, 1].min())/2

        if faces_max_radius is None:
            faces_max_radius = np.mean(clipped_hull_mesh.faces_radiuses)

        candidate_lid_size = (
                    max(faces_max_radius/2, 1.1*x_span),
                    max(faces_max_radius/2, 1.1*y_span),
                )
        # The size of the lid is at least the characteristic length of a face

        candidate_lid_mesh = mesh_rectangle(
                size=(candidate_lid_size[1], candidate_lid_size[0]),  # TODO Fix: Exchange x and y in mesh_rectangle
                faces_max_radius=faces_max_radius,
                center=(x_mean, y_mean, z),
                normal=(0.0, 0.0, -1.0),
                )

        candidate_lid_points = candidate_lid_mesh.vertices[:, 0:2]

        hull_faces = clipped_hull_mesh.vertices[clipped_hull_mesh.faces, 0:2]
        edges_of_hull_faces = hull_faces[:, [1, 2, 3, 0], :] - hull_faces[:, :, :]  # Vectors between two consecutive points in a face
        # edges_of_hull_faces.shape = (nb_full_faces, 4, 2)
        lid_points_in_local_coords = candidate_lid_points[:, np.newaxis, np.newaxis, :] - hull_faces[:, :, :]
        # lid_points_in_local_coords.shape = (nb_candidate_lid_points, nb_full_faces, 4, 2)
        side_of_hull_edges = (lid_points_in_local_coords[..., 0] * edges_of_hull_faces[..., 1]
                             - lid_points_in_local_coords[..., 1] * edges_of_hull_faces[..., 0])
        # side_of_hull_edges.shape = (nb_candidate_lid_points, nb_full_faces, 4)
        point_is_above_panel = np.all(side_of_hull_edges <= 0, axis=-1) | np.all(side_of_hull_edges >= 0, axis=-1)
        # point_is_above_panel.shape = (nb_candidate_lid_points, nb_full_faces)

        # For all point in candidate_lid_points, and for all edges of all faces of
        # the hull mesh, check on which side of the edge is the point by using a
        # cross product.
        # If a point on the same side of all edges of a face, then it is inside.

        nb_panels_below_point = np.sum(point_is_above_panel, axis=-1)
        needs_lid = (nb_panels_below_point % 2 == 1).nonzero()[0]

        lid_faces = candidate_lid_mesh.faces[np.all(np.isin(candidate_lid_mesh.faces, needs_lid), axis=-1), :]

        if name is None:
            name = "lid for {}".format(self.name)

        if len(lid_faces) == 0:
            return Mesh(None, None, name=name)

        lid_mesh = Mesh(candidate_lid_mesh.vertices, lid_faces, name=name)
        lid_mesh.heal_mesh()

        return lid_mesh

    @inplace_transformation
    def with_normal_vector_going_down(self):
        # For lid meshes for irregular frequencies removal
        if np.allclose(self.faces_normals[:, 2], np.ones((self.nb_faces,))):
            # The mesh is horizontal with normal vectors going up
            LOG.warning(f"Inverting the direction of the normal vectors of {self} to be downward.")
            self.faces = self.faces[:, ::-1]
        else:
            return self

    def _face_on_plane(self, i_face, plane):
        return (
                self.faces_centers[i_face, :] in plane
                and plane.is_orthogonal_to(self.faces_normals[i_face, :])
                )

    def extract_lid(self, plane=xOy_Plane):
        """
        Split the mesh into a mesh of the hull and a mesh of the lid.
        By default, the lid is composed of the horizontal faces on the z=0 plane.

        Parameters
        ----------
        plane: Plane
            The plane on which to look for lid faces.

        Returns
        -------
        2-ple of Mesh
            hull mesh and lid mesh
        """
        faces_on_plane = [i_face for i_face in range(self.nb_faces) if self._face_on_plane(i_face, plane)]
        lid_mesh = self.extract_faces(faces_on_plane)
        hull_mesh = self.extract_faces(list(set(range(self.nb_faces)) - set(faces_on_plane)))
        return hull_mesh, lid_mesh



import pytest
import numpy as np
from numpy.linalg import norm
import capytaine as cpt


sphere = cpt.mesh_sphere(radius=1)


def test_clipper():
    """Test clipping of mesh."""
    mesh = cpt.mesh_sphere(radius=5.0, resolution=(10, 5))
    aabb = mesh.axis_aligned_bbox

    mesh.keep_immersed_part(free_surface=0.0, water_depth=np.inf)
    assert np.allclose(mesh.axis_aligned_bbox, aabb[:5] + (0,))  # the last item of the tuple has changed

    mesh.keep_immersed_part(free_surface=0.0, water_depth=1.0)
    assert np.allclose(mesh.axis_aligned_bbox, aabb[:4] + (-1, 0,))  # the last item of the tuple has changed

    # With CollectionOfMeshes (AxialSymmetry)
    mesh = cpt.mesh_sphere(radius=5.0, resolution=(10, 5), axial_symmetry=True)
    aabb = mesh.merged().axis_aligned_bbox

    mesh.keep_immersed_part(free_surface=0.0, water_depth=np.inf)
    assert np.allclose(mesh.merged().axis_aligned_bbox, aabb[:5] + (0,))  # the last item of the tuple has changed

    mesh.keep_immersed_part(free_surface=0.0, water_depth=1.0)
    assert np.allclose(mesh.merged().axis_aligned_bbox, aabb[:4] + (-1, 0,))  # the last item of the tuple has changed

    # Check boundaries after clipping
    mesh = cpt.mesh_rectangle(size=(5,5), normal=(1,0,0))
    assert max([i[2] for i in mesh.immersed_part(free_surface=-1).vertices])<=-1
    assert max([i[2] for i in mesh.immersed_part(free_surface= 1).vertices])<= 1
    assert min([i[2] for i in mesh.immersed_part(free_surface=100, sea_bottom=-1).vertices])>=-1
    assert min([i[2] for i in mesh.immersed_part(free_surface=100, sea_bottom= 1).vertices])>= 1

    mesh = cpt.mesh_rectangle(size=(4,4), resolution=(1,1), normal=(1,0,0))
    tmp = list(mesh.clip(cpt.Plane(normal=(0,0.1,1),point=(0,0,-1)),inplace=False).vertices)
    tmp.sort(key=lambda x: x[2])
    tmp.sort(key=lambda x: x[1])
    assert np.allclose([i[2] for i in tmp], [-2, -0.8, -2, -1.2])


@pytest.mark.parametrize("size", [5, 6])
def test_clipper_indices(size):
    """Test clipped_mesh_faces_ids."""
    mesh = cpt.mesh_rectangle(size=(size, size), resolution=(size, size), center=(0, 0, 0))
    clipped_mesh = mesh.clipped(plane=cpt.Plane(point=(0, 0, 0), normal=(0, 0, 1)))
    faces_ids = clipped_mesh._clipping_data['faces_ids']

    assert clipped_mesh.nb_faces == len(faces_ids)
    assert all(norm(clipped_mesh.faces_centers[i] - mesh.faces_centers[face_id]) < 0.3
               for i, face_id in enumerate(faces_ids))


def test_clipper_corner_cases():
    mesh = sphere.translated_z(10.0)

    plane = cpt.Plane(point=(0, 0, 0), normal=(0, 0, 1))
    clipped_mesh = mesh.clip(plane, inplace=False)
    assert clipped_mesh == cpt.Mesh(None, None)  # Empty mesh

    plane = cpt.Plane(point=(0, 0, 0), normal=(0, 0, -1))
    clipped_mesh = mesh.clip(plane, inplace=False)
    assert clipped_mesh == mesh  # Unchanged mesh

    # Two distinct bodies
    two_spheres = cpt.Mesh.join_meshes(sphere.translated_z(10.0), sphere.translated_z(-10.0))
    plane = cpt.Plane(point=(0, 0, 0), normal=(0, 0, -1))
    one_sphere_remaining = two_spheres.clip(plane, inplace=False)
    assert one_sphere_remaining == sphere.translated_z(10.0)


def test_clipper_tolerance():
    mesh = cpt.mesh_vertical_cylinder(length=10.001, center=(0, 0, -5))
    mesh = mesh.immersed_part()
    np.testing.assert_allclose(mesh.vertices[:, 2].max(), 0.0, atol=1e-12)


def test_degenerate_faces():
    vertices = np.array([
        [-8.00000000e+00,  1.65358984e+00, -4.99999996e-02],
        [-8.00000000e+00,  1.65358984e+00,  5.00000003e-02],
        [-8.00000000e+00,  1.74019238e+00, -9.99999998e-02],
        [-8.00000000e+00,  1.74019238e+00, -1.78037182e-10],
        [-8.00000000e+00,  1.74019238e+00,  1.00000000e-01],
        [-8.00000000e+00,  1.82679492e+00, -5.00000002e-02],
        [-8.00000000e+00,  1.82679492e+00,  4.99999997e-02]
        ])
    faces = np.array([
        [5, 3, 6, 5],
        [5, 2, 3, 5],
        [3, 4, 6, 3],
        [3, 4, 6, 3],
        [2, 0, 3, 2],
        [0, 1, 3, 0],
        [1, 4, 3, 1]
        ])
    mesh = cpt.Mesh(vertices, faces)
    clipped_mesh = mesh.immersed_part()
    assert clipped_mesh.nb_faces == 4



# Copyright (C) 2017-2024 Matthieu Ancellin
# See LICENSE file at <https://github.com/capytaine/capytaine>
"""Solver for the BEM problem.

.. code-block:: python

    problem = RadiationProblem(...)
    result = BEMSolver(green_functions=..., engine=...).solve(problem)

"""

import os
import logging

import numpy as np
import pandas as pd

from datetime import datetime

from rich.progress import track

from capytaine.bem.problems_and_results import LinearPotentialFlowProblem, DiffractionProblem
from capytaine.green_functions.delhommeau import Delhommeau
from capytaine.bem.engines import BasicMatrixEngine
from capytaine.io.xarray import problems_from_dataset, assemble_dataset, kochin_data_array
from capytaine.tools.optional_imports import silently_import_optional_dependency
from capytaine.tools.lists_of_points import _normalize_points, _normalize_free_surface_points
from capytaine.tools.symbolic_multiplication import supporting_symbolic_multiplication
from capytaine.tools.timer import Timer

LOG = logging.getLogger(__name__)

class BEMSolver:
    """
    Solver for linear potential flow problems.

    Parameters
    ----------
    green_function: AbstractGreenFunction, optional
        Object handling the computation of the Green function.
        (default: :class:`~capytaine.green_function.delhommeau.Delhommeau`)
    engine: MatrixEngine, optional
        Object handling the building of matrices and the resolution of linear systems with these matrices.
        (default: :class:`~capytaine.bem.engines.BasicMatrixEngine`)
    method: string, optional
        select boundary integral equation used to solve the problems.
        Accepted values: "indirect" (as in e.g. Nemoh), "direct" (as in e.g. WAMIT)
        Default value: "indirect"

    Attributes
    ----------
    timer: dict[str, Timer]
        Storing the time spent on each subtasks of the resolution
    exportable_settings : dict
        Settings of the solver that can be saved to reinit the same solver later.
    """

    def __init__(self, *, green_function=None, engine=None, method="indirect"):
        self.green_function = Delhommeau() if green_function is None else green_function
        self.engine = BasicMatrixEngine() if engine is None else engine

        if method.lower() not in {"direct", "indirect"}:
            raise ValueError(f"Unrecognized method when initializing solver: {repr(method)}. Expected \"direct\" or \"indirect\".")
        self.method = method.lower()

        self.timer = {"Solve total": Timer(), "  Green function": Timer(), "  Linear solver": Timer()}

        self.solve = self.timer["Solve total"].wraps_function(self.solve)

        try:
            self.exportable_settings = {
                **self.green_function.exportable_settings,
                **self.engine.exportable_settings,
                "method": self.method,
            }
        except AttributeError:
            self.exportable_settings = {}

    def __str__(self):
        return f"BEMSolver(engine={self.engine}, green_function={self.green_function})"

    def __repr__(self):
        return self.__str__()

    def timer_summary(self):
        return pd.DataFrame([
            {
                "task": name,
                "total": self.timer[name].total,
                "nb_calls": self.timer[name].nb_timings,
                "mean": self.timer[name].mean
            } for name in self.timer]).set_index("task")

    def _repr_pretty_(self, p, cycle):
        p.text(self.__str__())

    @classmethod
    def from_exported_settings(settings):
        raise NotImplementedError

    def solve(self, problem, method=None, keep_details=True, _check_wavelength=True):
        """Solve the linear potential flow problem.

        Parameters
        ----------
        problem: LinearPotentialFlowProblem
            the problem to be solved
        method: string, optional
            select boundary integral equation used to solve the problem.
            It is recommended to set the method more globally when initializing the solver.
            If provided here, the value in argument of `solve` overrides the global one.
        keep_details: bool, optional
            if True, store the sources and the potential on the floating body in the output object
            (default: True)
        _check_wavelength: bool, optional (default: True)
            If True, the frequencies are compared to the mesh resolution and
            the estimated first irregular frequency to warn the user.

        Returns
        -------
        LinearPotentialFlowResult
            an object storing the problem data and its results
        """
        LOG.info("Solve %s.", problem)

        if _check_wavelength:
            self._check_wavelength_and_mesh_resolution([problem])
            self._check_wavelength_and_irregular_frequencies([problem])

        if isinstance(problem, DiffractionProblem) and float(problem.encounter_omega) in {0.0, np.inf}:
            raise ValueError("Diffraction problems at zero or infinite frequency are not defined")
            # This error used to be raised when initializing the problem.
            # It is now raised here, in order to be catchable by
            # _solve_and_catch_errors, such that batch resolution
            # can include this kind of problems without the full batch
            # failing.
            # Note that if this error was not raised here, the resolution
            # would still fail with a less explicit error message.

        if problem.forward_speed != 0.0:
            omega, wavenumber = problem.encounter_omega, problem.encounter_wavenumber
        else:
            omega, wavenumber = problem.omega, problem.wavenumber

        linear_solver = supporting_symbolic_multiplication(self.engine.linear_solver)
        method = method if method is not None else self.method
        if (method == 'direct'):
            if problem.forward_speed != 0.0:
                raise NotImplementedError("Direct solver is not able to solve problems with forward speed.")

            with self.timer["  Green function"]:
                S, D = self.engine.build_matrices(
                        problem.body.mesh_including_lid, problem.body.mesh_including_lid,
                        problem.free_surface, problem.water_depth, wavenumber,
                        self.green_function, adjoint_double_layer=False
                        )
            rhs = S @ problem.boundary_condition
            with self.timer["  Linear solver"]:
                potential = linear_solver(D, rhs)
            if not potential.shape == problem.boundary_condition.shape:
                raise ValueError(f"Error in linear solver of {self.engine}: the shape of the output ({potential.shape}) "
                                 f"does not match the expected shape ({problem.boundary_condition.shape})")
            pressure = 1j * omega * problem.rho * potential
            sources = None
        else:
            with self.timer["  Green function"]:
                S, K = self.engine.build_matrices(
                        problem.body.mesh_including_lid, problem.body.mesh_including_lid,
                        problem.free_surface, problem.water_depth, wavenumber,
                        self.green_function, adjoint_double_layer=True
                        )

            with self.timer["  Linear solver"]:
                sources = linear_solver(K, problem.boundary_condition)
            if not sources.shape == problem.boundary_condition.shape:
                raise ValueError(f"Error in linear solver of {self.engine}: the shape of the output ({sources.shape}) "
                                 f"does not match the expected shape ({problem.boundary_condition.shape})")
            potential = S @ sources
            pressure = 1j * omega * problem.rho * potential
            if problem.forward_speed != 0.0:
                result = problem.make_results_container(sources=sources)
                # Temporary result object to compute the ∇Φ term
                nabla_phi = self._compute_potential_gradient(problem.body.mesh_including_lid, result)
                pressure += problem.rho * problem.forward_speed * nabla_phi[:, 0]

        pressure_on_hull = pressure[problem.body.hull_mask]  # Discards pressure on lid if any
        forces = problem.body.integrate_pressure(pressure_on_hull)

        if not keep_details:
            result = problem.make_results_container(forces)
        else:
            result = problem.make_results_container(forces, sources, potential, pressure)

        LOG.debug("Done!")

        return result

    def _solve_and_catch_errors(self, problem, *args, **kwargs):
        """Same as BEMSolver.solve() but returns a
        FailedLinearPotentialFlowResult when the resolution failed."""
        try:
            res = self.solve(problem, *args, **kwargs)
        except Exception as e:
            LOG.info(f"Skipped {problem}\nbecause of {repr(e)}")
            res = problem.make_failed_results_container(e)
        return res

    def solve_all(self, problems, *, method=None, n_jobs=1, progress_bar=None, _check_wavelength=True, **kwargs):
        """Solve several problems.
        Optional keyword arguments are passed to `BEMSolver.solve`.

        Parameters
        ----------
        problems: list of LinearPotentialFlowProblem
            several problems to be solved
        method: string, optional
            select boundary integral equation used to solve the problems.
            It is recommended to set the method more globally when initializing the solver.
            If provided here, the value in argument of `solve_all` overrides the global one.
        n_jobs: int, optional (default: 1)
            the number of jobs to run in parallel using the optional dependency `joblib`
            By defaults: do not use joblib and solve sequentially.
        progress_bar: bool, optional
            Display a progress bar while solving.
            If no value is provided to this method directly,
            check whether the environment variable `CAPYTAINE_PROGRESS_BAR` is defined
            and otherwise default to True.
        _check_wavelength: bool, optional (default: True)
            If True, the frequencies are compared to the mesh resolution and
            the estimated first irregular frequency to warn the user.

        Returns
        -------
        list of LinearPotentialFlowResult
            the solved problems
        """
        if _check_wavelength:
            self._check_wavelength_and_mesh_resolution(problems)
            self._check_wavelength_and_irregular_frequencies(problems)

        if progress_bar is None:
            if "CAPYTAINE_PROGRESS_BAR" in os.environ:
                env_var = os.environ["CAPYTAINE_PROGRESS_BAR"].lower()
                if env_var in {'true', '1', 't'}:
                    progress_bar = True
                elif env_var in {'false', '0', 'f'}:
                    progress_bar = False
                else:
                    raise ValueError("Invalid value '{}' for the environment variable CAPYTAINE_PROGRESS_BAR.".format(os.environ["CAPYTAINE_PROGRESS_BAR"]))
            else:
                progress_bar = True

        if n_jobs == 1:  # force sequential resolution
            problems = sorted(problems)
            if progress_bar:
                problems = track(problems, total=len(problems), description="Solving BEM problems")
            results = [self._solve_and_catch_errors(pb, method=method, _check_wavelength=False, **kwargs) for pb in problems]
        else:
            joblib = silently_import_optional_dependency("joblib")
            if joblib is None:
                raise ImportError(f"Setting the `n_jobs` argument to {n_jobs} requires the missing optional dependency 'joblib'.")
            groups_of_problems = LinearPotentialFlowProblem._group_for_parallel_resolution(problems)
            parallel = joblib.Parallel(return_as="generator", n_jobs=n_jobs)
            groups_of_results = parallel(joblib.delayed(self.solve_all)(grp, method=method, n_jobs=1, progress_bar=False, _check_wavelength=False, **kwargs) for grp in groups_of_problems)
            if progress_bar:
                groups_of_results = track(groups_of_results,
                                          total=len(groups_of_problems),
                                          description=f"Solving BEM problems with {n_jobs} threads:")
            results = [res for grp in groups_of_results for res in grp]  # flatten the nested list
        LOG.info("Solver timer summary:\n%s", self.timer_summary())
        return results

    @staticmethod
    def _check_wavelength_and_mesh_resolution(problems):
        """Display a warning if some of the problems have a mesh resolution
        that might not be sufficient for the given wavelength."""
        LOG.debug("Check wavelength with mesh resolution.")
        risky_problems = [pb for pb in problems
                          if 0.0 < pb.wavelength < pb.body.minimal_computable_wavelength]
        nb_risky_problems = len(risky_problems)
        if nb_risky_problems == 1:
            pb = risky_problems[0]
            freq_type = risky_problems[0].provided_freq_type
            freq = pb.__getattribute__(freq_type)
            LOG.warning(f"Mesh resolution for {pb}:\n"
                        f"The resolution of the mesh of the body {pb.body.__short_str__()} might "
                        f"be insufficient for {freq_type}={freq}.\n"
                        "This warning appears because the largest panel of this mesh "
                        f"has radius {pb.body.mesh.faces_radiuses.max():.3f} > wavelength/8."
                        )
        elif nb_risky_problems > 1:
            freq_type = risky_problems[0].provided_freq_type
            freqs = np.array([float(pb.__getattribute__(freq_type)) for pb in risky_problems])
            LOG.warning(f"Mesh resolution for {nb_risky_problems} problems:\n"
                        "The resolution of the mesh might be insufficient "
                        f"for {freq_type} ranging from {freqs.min():.3f} to {freqs.max():.3f}.\n"
                        "This warning appears when the largest panel of this mesh "
                        "has radius > wavelength/8."
                        )

    @staticmethod
    def _check_wavelength_and_irregular_frequencies(problems):
        """Display a warning if some of the problems might encounter irregular frequencies."""
        LOG.debug("Check wavelength with estimated irregular frequency.")
        risky_problems = [pb for pb in problems
                          if pb.free_surface != np.inf and
                          pb.body.first_irregular_frequency_estimate(g=pb.g) < pb.omega < np.inf]
        nb_risky_problems = len(risky_problems)
        if nb_risky_problems >= 1:
            if any(pb.body.lid_mesh is None for pb in problems):
                recommendation = "Setting a lid for the floating body is recommended."
            else:
                recommendation = "The lid might need to be closer to the free surface."
            if nb_risky_problems == 1:
                pb = risky_problems[0]
                freq_type = risky_problems[0].provided_freq_type
                freq = pb.__getattribute__(freq_type)
                LOG.warning(f"Irregular frequencies for {pb}:\n"
                            f"The body {pb.body.__short_str__()} might display irregular frequencies "
                            f"for {freq_type}={freq}.\n"
                            + recommendation
                            )
            elif nb_risky_problems > 1:
                freq_type = risky_problems[0].provided_freq_type
                freqs = np.array([float(pb.__getattribute__(freq_type)) for pb in risky_problems])
                LOG.warning(f"Irregular frequencies for {nb_risky_problems} problems:\n"
                            "Irregular frequencies might be encountered "
                            f"for {freq_type} ranging from {freqs.min():.3f} to {freqs.max():.3f}.\n"
                            + recommendation
                            )

    def fill_dataset(self, dataset, bodies, *, method=None, n_jobs=1, _check_wavelength=True, progress_bar=None, **kwargs):
        """Solve a set of problems defined by the coordinates of an xarray dataset.

        Parameters
        ----------
        dataset : xarray Dataset
            dataset containing the problems parameters: frequency, radiating_dof, water_depth, ...
        bodies : FloatingBody or list of FloatingBody
            The body or bodies involved in the problems
            They should all have different names.
        method: string, optional
            select boundary integral equation used to solve the problems.
            It is recommended to set the method more globally when initializing the solver.
            If provided here, the value in argument of `fill_dataset` overrides the global one.
        n_jobs: int, optional (default: 1)
            the number of jobs to run in parallel using the optional dependency `joblib`
            By defaults: do not use joblib and solve sequentially.
        progress_bar: bool, optional
            Display a progress bar while solving.
            If no value is provided to this method directly,
            check whether the environment variable `CAPYTAINE_PROGRESS_BAR` is defined
            and otherwise default to True.
        _check_wavelength: bool, optional (default: True)
            If True, the frequencies are compared to the mesh resolution and
            the estimated first irregular frequency to warn the user.

        Returns
        -------
        xarray Dataset
        """
        attrs = {'start_of_computation': datetime.now().isoformat(),
                 **self.exportable_settings}
        if method is not None:  # Overrides the method in self.exportable_settings
            attrs["method"] = method
        problems = problems_from_dataset(dataset, bodies)
        if 'theta' in dataset.coords:
            results = self.solve_all(problems, keep_details=True, method=method, n_jobs=n_jobs, _check_wavelength=_check_wavelength, progress_bar=progress_bar)
            kochin = kochin_data_array(results, dataset.coords['theta'])
            dataset = assemble_dataset(results, attrs=attrs, **kwargs)
            dataset.update(kochin)
        else:
            results = self.solve_all(problems, keep_details=False, method=method, n_jobs=n_jobs, _check_wavelength=_check_wavelength, progress_bar=progress_bar)
            dataset = assemble_dataset(results, attrs=attrs, **kwargs)
        return dataset


    def compute_potential(self, points, result):
        """Compute the value of the potential at given points for a previously solved potential flow problem.

        Parameters
        ----------
        points: array of shape (3,) or (N, 3), or 3-ple of arrays returned by meshgrid, or MeshLike object
            Coordinates of the point(s) at which the potential should be computed
        result: LinearPotentialFlowResult
            The return of the BEM solver

        Returns
        -------
        complex-valued array of shape (1,) or (N,) or (nx, ny, nz) or (mesh.nb_faces,) depending of the kind of input
            The value of the potential at the points

        Raises
        ------
        Exception: if the :code:`LinearPotentialFlowResult` object given as input does not contain the source distribution.
        """
        points, output_shape = _normalize_points(points, keep_mesh=True)
        if result.sources is None:
            raise Exception(f"""The values of the sources of {result} cannot been found.
            They probably have not been stored by the solver because the option keep_details=True have not been set or the direct method has been used.
            Please re-run the resolution with the indirect method and keep_details=True.""")

        with self.timer["  Green function"]:
            S, _ = self.green_function.evaluate(points, result.body.mesh_including_lid, result.free_surface, result.water_depth, result.encounter_wavenumber)
        potential = S @ result.sources  # Sum the contributions of all panels in the mesh
        return potential.reshape(output_shape)

    def _compute_potential_gradient(self, points, result):
        points, output_shape = _normalize_points(points, keep_mesh=True)

        if result.sources is None:
            raise Exception(f"""The values of the sources of {result} cannot been found.
            They probably have not been stored by the solver because the option keep_details=True have not been set.
            Please re-run the resolution with this option.""")

        with self.timer["  Green function"]:
            _, gradG = self.green_function.evaluate(points, result.body.mesh_including_lid, result.free_surface, result.water_depth, result.encounter_wavenumber,
                                                early_dot_product=False)
        velocities = np.einsum('ijk,j->ik', gradG, result.sources)  # Sum the contributions of all panels in the mesh
        return velocities.reshape((*output_shape, 3))

    def compute_velocity(self, points, result):
        """Compute the value of the velocity vector at given points for a previously solved potential flow problem.

        Parameters
        ----------
        points: array of shape (3,) or (N, 3), or 3-ple of arrays returned by meshgrid, or MeshLike object
            Coordinates of the point(s) at which the velocity should be computed
        result: LinearPotentialFlowResult
            The return of the BEM solver

        Returns
        -------
        complex-valued array of shape (3,) or (N,, 3) or (nx, ny, nz, 3) or (mesh.nb_faces, 3) depending of the kind of input
            The value of the velocity at the points

        Raises
        ------
        Exception: if the :code:`LinearPotentialFlowResult` object given as input does not contain the source distribution.
        """
        nabla_phi = self._compute_potential_gradient(points, result)
        if result.forward_speed != 0.0:
            nabla_phi[..., 0] -= result.forward_speed
        return nabla_phi

    def compute_pressure(self, points, result):
        """Compute the value of the pressure at given points for a previously solved potential flow problem.

        Parameters
        ----------
        points: array of shape (3,) or (N, 3), or 3-ple of arrays returned by meshgrid, or MeshLike object
            Coordinates of the point(s) at which the pressure should be computed
        result: LinearPotentialFlowResult
            The return of the BEM solver

        Returns
        -------
        complex-valued array of shape (1,) or (N,) or (nx, ny, nz) or (mesh.nb_faces,) depending of the kind of input
            The value of the pressure at the points

        Raises
        ------
        Exception: if the :code:`LinearPotentialFlowResult` object given as input does not contain the source distribution.
        """
        if result.forward_speed != 0:
            pressure = 1j * result.encounter_omega * result.rho * self.compute_potential(points, result)
            nabla_phi = self._compute_potential_gradient(points, result)
            pressure += result.rho * result.forward_speed * nabla_phi[..., 0]
        else:
            pressure = 1j * result.omega * result.rho * self.compute_potential(points, result)
        return pressure


    def compute_free_surface_elevation(self, points, result):
        """Compute the value of the free surface elevation at given points for a previously solved potential flow problem.

        Parameters
        ----------
        points: array of shape (2,) or (N, 2), or 2-ple of arrays returned by meshgrid, or MeshLike object
            Coordinates of the point(s) at which the free surface elevation should be computed
        result: LinearPotentialFlowResult
            The return of the BEM solver

        Returns
        -------
        complex-valued array of shape (1,) or (N,) or (nx, ny, nz) or (mesh.nb_faces,) depending of the kind of input
            The value of the free surface elevation at the points

        Raises
        ------
        Exception: if the :code:`LinearPotentialFlowResult` object given as input does not contain the source distribution.
        """
        points, output_shape = _normalize_free_surface_points(points, keep_mesh=True)

        if result.forward_speed != 0:
            fs_elevation = -1/result.g * (-1j*result.encounter_omega) * self.compute_potential(points, result)
            nabla_phi = self._compute_potential_gradient(points, result)
            fs_elevation += -1/result.g * result.forward_speed * nabla_phi[..., 0]
        else:
            fs_elevation = -1/result.g * (-1j*result.omega) * self.compute_potential(points, result)

        return fs_elevation.reshape(output_shape)


    ## Legacy

    def get_potential_on_mesh(self, result, mesh, chunk_size=50):
        """Compute the potential on a mesh for the potential field of a previously solved problem.
        Since the interaction matrix does not need to be computed in full to compute the matrix-vector product,
        only a few lines are evaluated at a time to reduce the memory cost of the operation.

        The newer method :code:`compute_potential` should be preferred in the future.

        Parameters
        ----------
        result : LinearPotentialFlowResult
            the return of the BEM solver
        mesh : MeshLike
            a mesh
        chunk_size: int, optional
            Number of lines to compute in the matrix.
            (legacy, should be passed as an engine setting instead).

        Returns
        -------
        array of shape (mesh.nb_faces,)
            potential on the faces of the mesh

        Raises
        ------
        Exception: if the :code:`Result` object given as input does not contain the source distribution.
        """
        LOG.info(f"Compute potential on {mesh.name} for {result}.")

        if result.sources is None:
            raise Exception(f"""The values of the sources of {result} cannot been found.
            They probably have not been stored by the solver because the option keep_details=True have not been set or the direct method has been used.
            Please re-run the resolution with the indirect method and keep_details=True.""")

        if chunk_size > mesh.nb_faces:
            S = self.engine.build_S_matrix(
                mesh,
                result.body.mesh_including_lid,
                result.free_surface, result.water_depth, result.wavenumber,
                self.green_function
            )
            phi = S @ result.sources

        else:
            phi = np.empty((mesh.nb_faces,), dtype=np.complex128)
            for i in range(0, mesh.nb_faces, chunk_size):
                faces_to_extract = list(range(i, min(i+chunk_size, mesh.nb_faces)))
                S = self.engine.build_S_matrix(
                    mesh.extract_faces(faces_to_extract),
                    result.body.mesh_including_lid,
                    result.free_surface, result.water_depth, result.wavenumber,
                    self.green_function
                )
                phi[i:i+chunk_size] = S @ result.sources

        LOG.debug(f"Done computing potential on {mesh.name} for {result}.")

        return phi

    def get_free_surface_elevation(self, result, free_surface, keep_details=False):
        """Compute the elevation of the free surface on a mesh for a previously solved problem.

        The newer method :code:`compute_free_surface_elevation` should be preferred in the future.

        Parameters
        ----------
        result : LinearPotentialFlowResult
            the return of the solver
        free_surface : FreeSurface
            a meshed free surface
        keep_details : bool, optional
            if True, keep the free surface elevation in the LinearPotentialFlowResult (default:False)

        Returns
        -------
        array of shape (free_surface.nb_faces,)
            the free surface elevation on each faces of the meshed free surface

        Raises
        ------
        Exception: if the :code:`Result` object given as input does not contain the source distribution.
        """
        if result.forward_speed != 0.0:
            raise NotImplementedError("For free surface elevation with forward speed, please use the `compute_free_surface_elevation` method.")

        fs_elevation = 1j*result.omega/result.g * self.get_potential_on_mesh(result, free_surface.mesh)
        if keep_details:
            result.fs_elevation[free_surface] = fs_elevation
        return fs_elevation



"""Tests for the mesh import of mesh from NEMOH file format"""
from unittest import mock

import capytaine as cpt
from capytaine.io.mesh_loaders import load_MAR


def parse_mar(file_content):
    # Workaround to avoid actually writing/reading a file on disk
    with mock.patch('capytaine.io.mesh_loaders._check_file', lambda foo: None):
        with mock.patch('capytaine.io.mesh_loaders.open', mock.mock_open(read_data=file_content)):
            return load_MAR("mocked/filename")


def test_mar_no_symmetry():
    mesh = parse_mar(
            """2 0
            1 0.0 0.0 -1.0
            2 1.0 0.0 -1.0
            3 1.0 1.0 -1.0
            4 0.0 1.0 -1.0
            5 1.0 0.0 -1.0
            6 2.0 0.0 -1.0
            7 2.0 1.0 -1.0
            8 1.0 1.0 -1.0
            0 0 0 0
            1 2 3 4
            5 6 7 8
            0 0 0 0
            """)
    assert isinstance(mesh, cpt.Mesh)
    assert mesh.nb_faces == 2


def test_mar_symmetry():
    mesh = parse_mar(
            """2 1
            1 0.0 0.0 -1.0
            2 1.0 0.0 -1.0
            3 1.0 1.0 -1.0
            4 0.0 1.0 -1.0
            5 1.0 0.0 -1.0
            6 2.0 0.0 -1.0
            7 2.0 1.0 -1.0
            8 1.0 1.0 -1.0
            0 0 0 0
            1 2 3 4
            5 6 7 8
            0 0 0 0
            """)
    assert isinstance(mesh, cpt.ReflectionSymmetricMesh)
    assert mesh.plane == cpt.xOz_Plane
    assert mesh.nb_faces == 4



import numpy as np
from pathlib import Path
import xarray as xr

import capytaine as cpt

# --- Parameters ---
output_dir = Path("outputs")
output_dir.mkdir(exist_ok=True)

# --- Create mesh and set up body ---
mesh = cpt.mesh_sphere()
dofs = cpt.rigid_body_dofs(rotation_center=(0, 0, 0))
full_body = cpt.FloatingBody(mesh, dofs)
full_body.center_of_mass = np.copy(full_body.mesh.center_of_mass_of_nodes)
immersed_body = full_body.immersed_part()
immersed_body.compute_hydrostatics()

# --- Define parameter grid ---
omegas = np.concatenate(([0], np.linspace(0.1, 1.0, 10), [np.inf]))  # 0, 5 values, inf
wave_directions = np.linspace(0, np.pi, 3)  # 3 directions: 0, pi/2, pi

test_matrix = xr.Dataset(
    {
        "omega": omegas,
        "wave_direction": wave_directions,
        "radiating_dof": list(immersed_body.dofs),
        "water_depth": [np.inf],
        "rho": [1025],
    }
)

# --- Run simulation ---
solver = cpt.BEMSolver()
dataset = solver.fill_dataset(test_matrix, immersed_body)


cpt.export_dataset(output_dir / "test_boat200.nc", dataset)  # Format is deduced from filename suffix.
cpt.export_dataset(output_dir / "wamit_data", dataset, format="wamit")
(output_dir / 'nemoh_data').mkdir(exist_ok=True)
cpt.export_dataset(output_dir / "nemoh_data", dataset, format="nemoh")

print("Exported files:")
for file in output_dir.glob("**/*"):
    print(file)



#!/usr/bin/env python
# coding: utf-8
"""Tests for the computation of the Green function using Fortran routines."""

import pytest
import numpy as np
import capytaine as cpt

import capytaine.green_functions.libs.Delhommeau_float64 as fortran_core

# #----------------------------------------------------------------------------------#
# gf = cpt.Delhommeau()
#
# legacy_grid = gf.fortran_core.constants.legacy_grid
# scaled_nemoh3_grid = gf.fortran_core.constants.scaled_nemoh3_grid
# '''
#
# for nr in [1000, 1500, 2000]:
#     r_range = gf.fortran_core.delhommeau_integrals.default_r_spacing(nr, 100.0, 1)
#     print(r_range[(r_range > 0.1) & (r_range < 1.0)])  # Print points between 0.1 and 1
#
# '''
# zmin = [-100,-251,-400]
# @pytest.mark.parametrize("zmin", zmin)
# def test_green_functions_z_spacing_zmin(zmin):
#     """Test default_z_spacing"""
#
#     nz = 124
#     zRange = gf.fortran_core.delhommeau_integrals.default_z_spacing(nz, zmin, scaled_nemoh3_grid)
#
#     dz = []
#     for i in range(1,len(zRange)):
#         dz.append(np.log10(abs(zRange[i]))-log10(abs(zRange[i-1])))
#     dzComp = dz[-1]*np.ones_like(dz)
#
#     assert abs(zmin-zRange[-1]) < 1e-5
#     assert all([abs(a - b) < 1e-5 for a, b in zip(dz, dzComp)])
#
#
# nz = [64,124,248]
# @pytest.mark.parametrize("nz", nz)
# def test_green_functions_z_spacing_nz(nz):
#     """Test default_z_spacing"""
#     zmin = -251
#     zRange = gf.fortran_core.delhommeau_integrals.default_z_spacing(nz,zmin, scaled_nemoh3_grid)
#
#     dz = []
#     for i in range(1,len(zRange)):
#         dz.append(np.log10(abs(zRange[i]))-log10(abs(zRange[i-1])))
#     dzComp = dz[-1]*np.ones_like(dz)
#
#     assert abs(zmin-zRange[-1]) < 1e-5
#     assert all([abs(a - b) < 1e-5 for a, b in zip(dz, dzComp)])
#
#
# nr = [100,328,480]
# @pytest.mark.parametrize("nr", nr)
# def test_green_functions_r_spacing_legacy_nr(nr):
#     """Test default_r_spacing legacy model"""
#
#     rmax = 4/3+abs(nr-32)/3
#     rRange = gf.fortran_core.delhommeau_integrals.default_r_spacing(nr, rmax, legacy_grid)
#
#     r_cal = []
#     for i in range(1,nr+1):
#         r_cal.append(min(10**((i-1)/5-6),4/3+abs(i-32)/3))
#
#     assert all([abs(a - b) < 1e-4 for a, b in zip(rRange, r_cal)])
#
#
# rmax = [50,100,200]
# @pytest.mark.parametrize("rmax", rmax)
# def test_green_functions_r_spacing_Nemoh_rmax(rmax):
#     """Test default_r_spacing Nemoh model"""
#
#     nr = 676
#     c = 81
#     r_logSpace = -np.log(10.0**(-10))/(c - 1.0)
#     rRange = gf.fortran_core.delhommeau_integrals.default_r_spacing(nr, rmax, scaled_nemoh3_grid)
#
#     r_cal = []
#     for i in range(1,nr+1):
#         if (i < c):
#             r_cal.append(10**(-10)*np.exp(i*r_logSpace))
#         else:
#             r_cal.append((rmax-1)/(nr-c)*(i-c)+1)
#
#     assert all([abs(a - b) < 1e-4 for a, b in zip(rRange, r_cal)])
#
#
# nr = [676,1352]
# @pytest.mark.parametrize("nr", nr)
# def test_green_functions_r_spacing_Nemoh_nr(nr):
#     """Test default_r_spacing Nemoh model"""
#
#     nr0 = 676
#     c = round(nr/nr0*81)
#     rmax = 100
#     r_logSpace = -np.log(10.0**(-10))/(c - 1.0)
#     rRange = gf.fortran_core.delhommeau_integrals.default_r_spacing(nr, rmax, scaled_nemoh3_grid)
#
#     r_cal = []
#     for i in range(1,nr+1):
#         if (i < c):
#             r_cal.append(10**(-10)*np.exp(i*r_logSpace))
#         else:
#             r_cal.append((rmax-1)/(nr-c)*(i-c)+1)
#
#     assert all([abs(a - b) < 1e-4 for a, b in zip(rRange, r_cal)])


@pytest.mark.parametrize("params", [
        (324, 100.0, fortran_core.constants.legacy_grid),
        (676, 100.0, fortran_core.constants.scaled_nemoh3_grid),
        (600, 120.0, fortran_core.constants.scaled_nemoh3_grid),
    ])
def test_r_range_inversion(params):
    nr, rmax, grid_shape = params
    r_range = fortran_core.delhommeau_integrals.default_r_spacing(nr, rmax, grid_shape)
    assert np.all(r_range[:-1] < r_range[1:])  # Check that array is sorted and strictly monotone
    indices = np.array([fortran_core.delhommeau_integrals.nearest_r_index(r, r_range, grid_shape) for r in r_range])
    if grid_shape == fortran_core.constants.legacy_grid:
        assert np.all(np.abs(indices - np.arange(1, nr+1)) <= 1)  # Legacy grid is a bit off, but we'll not fix it for now
    else:
        assert np.all(np.abs(indices - np.arange(1, nr+1)) == 0)


@pytest.mark.parametrize("params", [
        (46, -16.0, fortran_core.constants.legacy_grid),
        (372, -251.0, fortran_core.constants.scaled_nemoh3_grid),
    ])
def test_z_range_inversion(params):
    nz, zmin, grid_shape = params
    z_range = fortran_core.delhommeau_integrals.default_z_spacing(nz, zmin, grid_shape)
    assert np.all(z_range[:-1] > z_range[1:])  # Check that array is sorted and strictly monotone
    indices = np.array([fortran_core.delhommeau_integrals.nearest_z_index(z, z_range, grid_shape) for z in z_range])
    if grid_shape == fortran_core.constants.legacy_grid:
        assert np.all(np.abs(indices - np.arange(1, nz+1)) <= 1)
    else:
        assert np.all(np.abs(indices - np.arange(1, nz+1)) == 0)
